import logging
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, Optional

from ..writer import LogWriter
from .base import BaseContainer
from .types import Entity

logger = logging.getLogger("MaximSDK")


@dataclass
class ToolCallConfig:
    id: str
    name: str
    description: str
    args: str
    tags: Optional[Dict[str, str]] = None


@dataclass
class ToolCallError:
    message: str
    code: Optional[str] = None
    type: Optional[str] = None


class ToolCall(BaseContainer):
    def __init__(self, config: ToolCallConfig, writer: LogWriter):
        super().__init__(Entity.TOOL_CALL, config.__dict__, writer)
        self._id = config.id
        self._name = config.name
        self.args = config.args
        self.description = config.description
        self.tags = config.tags

    @staticmethod
    def result_(writer: LogWriter, id: str, result: str):
        BaseContainer._commit_(
            writer, Entity.TOOL_CALL, id, "result", {"result": result}
        )
        BaseContainer._end_(
            writer,
            Entity.TOOL_CALL,
            id,
            {
                "endTimestamp": datetime.now(timezone.utc),
            },
        )

    def result(self, result: str):
        self._commit("result", {"result": result})
        self.end()

    def error(self, error: ToolCallError):
        self._commit("error", {"error": error})
        self.end()

    @staticmethod
    def error_(writer: LogWriter, id: str, error: ToolCallError):
        BaseContainer._commit_(writer, Entity.TOOL_CALL, id, "error", {"error": error})
        BaseContainer._end_(
            writer,
            Entity.TOOL_CALL,
            id,
            {
                "endTimestamp": datetime.now(timezone.utc),
            },
        )

    def data(self) -> Dict[str, Any]:
        base_data = super().data()
        return {
            **base_data,
            "name": self._name,
            "description": self.description,
            "args": self.args,
        }



import logging

logger = logging.getLogger("MaximSDK")

try:
    import langchain

    LANGCHAIN_AVAILABLE = True
except ImportError:
    LANGCHAIN_AVAILABLE = False

if not LANGCHAIN_AVAILABLE:
    logger.error(
        "LangChain is not available. You can't use MaximLangchainTracer.")


from .tracer import *
from .utils import *



import json
from dataclasses import asdict, dataclass
from datetime import date, datetime
from enum import Enum
from types import SimpleNamespace
from typing import Any, Dict, Optional


class Entity(Enum):
    SESSION = "session"
    TRACE = "trace"
    SPAN = "span"
    TOOL_CALL = "tool_call"
    GENERATION = "generation"
    FEEDBACK = "feedback"
    RETRIEVAL = "retrieval"


class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        return super().default(obj)


class CommitLog:
    def __init__(self, entity: Entity, entity_id: str, action: str, data: Optional[Dict[str, Any]] = None):
        self.entity = entity
        self.entity_id = entity_id
        self.action = action
        self.data = data

    def serialize(self, custom_data: Optional[Dict[str,Any]] = None) -> str:
        if custom_data is not None:
            if self.data is None:
                self.data = {}
            self.data.update(custom_data)
        return f"{self.entity.value}{{id={self.entity_id},action={self.action},data={json.dumps(self.data,cls=DateTimeEncoder)}}}"


def object_to_dict(obj):
    if isinstance(obj, dict):
        return {k: object_to_dict(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [object_to_dict(item) for item in obj]
    elif isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    elif isinstance(obj, (datetime, date)):
        return obj.isoformat()
    elif hasattr(obj, "__dict__"):
        return object_to_dict(obj.__dict__)
    elif isinstance(obj, SimpleNamespace):
        return object_to_dict(vars(obj))
    else:
        return str(obj)

@dataclass
class GenerationError:
    message: str
    code: Optional[str] = None
    type: Optional[str] = None



import logging
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional
from uuid import uuid4

from ..parsers.generation_parser import parse_model_parameters, parse_result
from ..writer import LogWriter
from .base import BaseContainer
from .types import Entity, GenerationError, object_to_dict

logger = logging.getLogger("MaximSDK")


@dataclass
class GenerationConfig:
    id: str
    provider: str
    model: str
    messages: Optional[List[Any]] = field(default_factory=list)
    model_parameters: Dict[str, Any] = field(default_factory=dict)
    span_id: Optional[str] = None
    name: Optional[str] = None
    maxim_prompt_id: Optional[str] = None
    tags: Optional[Dict[str, str]] = None


valid_providers = [
    "openai",
    "azure",
    "anthropic",
    "huggingface",
    "together",
    "google",
    "groq",
    "bedrock",
    "cohere",
    "unknown",
]


class Generation(BaseContainer):
    def __init__(self, config: GenerationConfig, writer: LogWriter):
        super().__init__(Entity.GENERATION, config.__dict__, writer)
        self.model = config.model
        self.maxim_prompt_id = config.maxim_prompt_id
        self.messages = []
        self.provider = config.provider
        self.provider = config.provider.lower()
        if self.provider not in valid_providers:
            self.provider = "unknown"
        if config.messages is not None:
            self.messages.extend(config.messages)
        self.model_parameters = parse_model_parameters(config.model_parameters)

    def _set_provider(self, writer: LogWriter, id: str, provider: str):
        if provider not in valid_providers:
            raise ValueError(
                f"Invalid provider: {self.provider}. Must be one of {', '.join(valid_providers)}."
            )
        BaseContainer._commit_(
            writer, Entity.GENERATION, id, "update", {"provider": provider}
        )

    def set_provider(self, provider: str):
        if provider not in valid_providers:
            raise ValueError(
                f"Invalid provider: {self.provider}. Must be one of {', '.join(valid_providers)}."
            )
        self.provider = provider

    @staticmethod
    def set_model_(writer: LogWriter, id: str, model: str):
        BaseContainer._commit_(
            writer, Entity.GENERATION, id, "update", {"model": model}
        )

    def set_model(self, model: str):
        self.model = model
        self._commit("update", {"model": model})

    @staticmethod
    def add_message_(writer: LogWriter, id: str, message: Any):
        BaseContainer._commit_(
            writer, Entity.GENERATION, id, "update", {"messages": [message]}
        )

    def add_message(self, message: Any):
        self.messages.append(message)
        self._commit("update", {"messages": [message]})

    @staticmethod
    def set_model_parameters_(
        writer: LogWriter, id: str, model_parameters: Dict[str, Any]
    ):
        BaseContainer._commit_(
            writer,
            Entity.GENERATION,
            id,
            "update",
            {"model_parameters": model_parameters},
        )

    def set_model_parameters(self, model_parameters: Dict[str, Any]):
        self.model_parameters = model_parameters
        self._commit("update", {"model_parameters": model_parameters})

    @staticmethod
    def result_(writer: LogWriter, id: str, result: Any):
        try:
            # Checking the type
            result = Generation.convert_result(result)
            # Validating the result
            parse_result(result)
            BaseContainer._commit_(
                writer, Entity.GENERATION, id, "result", {"result": result}
            )
            BaseContainer._end_(
                writer,
                Entity.GENERATION,
                id,
                {
                    "endTimestamp": datetime.now(timezone.utc),
                },
            )
        except ValueError as e:
            logger.error(
                "Invalid result. You can pass OpenAI/Azure ChatCompletion or Langchain LLMResult,AIMessage: {e}"
            )
            raise ValueError(
                f"Invalid result. You can pass OpenAI/Azure ChatCompletion or Langchain LLMResult,AIMessage: {e}"
            )

    @staticmethod
    def end_(writer: LogWriter, id: str, data: Optional[Dict[str, Any]] = None):
        if data is None:
            data = {}
        BaseContainer._end_(
            writer,
            Entity.GENERATION,
            id,
            {
                "endTimestamp": datetime.now(timezone.utc),
                **data,
            },
        )

    @staticmethod
    def add_tag_(writer: LogWriter, id: str, key: str, value: str):
        BaseContainer._add_tag_(writer, Entity.GENERATION, id, key, value)

    @staticmethod
    def convert_chat_completion(chat_completion: dict):
        return {
            "id": chat_completion.get("id", str(uuid4())),
            "created": chat_completion.get("created", datetime.now(timezone.utc)),
            "choices": [
                {
                    "index": choice.get("index", 0),
                    "message": {
                        "role": choice.get("message").get("role", "assistant"),
                        "content": choice.get("message").get("content", ""),
                        "tool_calls": choice.get("message").get("tool_calls", None),
                        "function_calls": choice.get("message").get(
                            "function_calls", None
                        ),
                    },
                    "finish_reason": choice.get("finish_reason", None),
                    "logprobs": choice.get("logprobs", None),
                }
                for choice in chat_completion.get("choices", [])
            ],
            "usage": chat_completion.get("usage", {}),
        }

    @staticmethod
    def convert_result(result: Any):
        try:
            parse_result(result)
            return result
        except Exception as err:
            result_dict = object_to_dict(result)
            if isinstance(result_dict, dict):
                # Checking if its Azure or OpenAI result
                if (
                    "object" in result_dict
                    and result_dict["object"] == "chat.completion"
                ):
                    return Generation.convert_chat_completion(result_dict)
                elif (
                    "object" in result_dict
                    and result_dict["object"] == "text.completion"
                ):
                    raise ValueError("Text completion is not yet supported.")
                elif isinstance(result, object):
                    try:
                        from langchain_core.messages import AIMessage
                        from langchain_core.outputs import LLMResult

                        from ..langchain.utils import (
                            parse_base_message_to_maxim_generation,
                            parse_langchain_llm_result)
                        if isinstance(result, AIMessage):
                            return parse_base_message_to_maxim_generation(result)
                        elif isinstance(result, LLMResult):
                            return parse_langchain_llm_result(result)
                    except ImportError:
                        pass
            return result

    def result(self, result: Any):
        try:
            # Checking the type
            result = Generation.convert_result(result)
            # Validating the result
            parse_result(result)
            # Logging the result
            self._commit("result", {"result": result})
            self.end()
        except ValueError as e:
            logger.error(
                f"Invalid result. You can pass OpenAI/Azure ChatCompletion or Langchain LLMResult,AIMessage: {e}"
            )
            raise ValueError(
                f"Invalid result. You can pass OpenAI/Azure ChatCompletion or Langchain LLMResult,AIMessage: {e}"
            )

    def error(self, error: GenerationError):
        if not error.code:
            error.code = ""
        if not error.type:
            error.type = ""
        self._commit(
            "result",
            {
                "result": {
                    "error": {
                        "message": error.message,
                        "code": error.code,
                        "type": error.type,
                    },
                    "id": str(uuid4()),
                }
            },
        )
        self.end()

    @staticmethod
    def error_(writer: LogWriter, id: str, error: GenerationError):
        if not error.code:
            error.code = ""
        if not error.type:
            error.type = ""
        BaseContainer._commit_(
            writer,
            Entity.GENERATION,
            id,
            "result",
            {
                "result": {
                    "error": {
                        "message": error.message,
                        "code": error.code,
                        "type": error.type,
                    },
                    "id": str(uuid4()),
                }
            },
        )
        BaseContainer._end_(
            writer,
            Entity.GENERATION,
            id,
            {
                "endTimestamp": datetime.now(timezone.utc),
            },
        )

    def data(self) -> Dict[str, Any]:
        base_data = super().data()
        return {
            **base_data,
            "model": self.model,
            "provider": self.provider,
            "maximPromptId": self.maxim_prompt_id,
            "messages": self.messages,
            "modelParameters": self.model_parameters,
        }



import asyncio
from contextvars import ContextVar
from functools import wraps
from typing import Callable, Optional
from uuid import uuid4

from ..logger.components.trace import Trace, TraceConfig
from ..logger.logger import Logger, Trace, TraceConfig

_maxim_ctx_var_trace: ContextVar[Optional[Trace]] = ContextVar(
    "maxim_ctx_trace", default=None
)

_maxim_ctx_var_logger: ContextVar[Optional[Logger]] = ContextVar(
    "maxim_ctx_logger", default=None
)


def current_logger() -> Optional[Logger]:
    try:
        return _maxim_ctx_var_logger.get()
    except LookupError:
        return None


def current_trace() -> Optional[Trace]:
    try:
        return _maxim_ctx_var_trace.get()
    except LookupError:
        return None


def trace(
    logger: Logger,
    id: Optional[str] or Optional[Callable] = None,
    sessionId: Optional[str] or Optional[Callable] = None,
    name: Optional[str] = None,
    tags: Optional[dict] = None,
):
    def decorator(func):
        if asyncio.iscoroutinefunction(func):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                actual_trace_id = id() if callable(id) else id
                if actual_trace_id is None:
                    actual_trace_id = str(uuid4())
                trace_config = TraceConfig(id=actual_trace_id, name=name, tags=tags)
                # If session_id is present, we will attach that to the trace
                actual_session_id = sessionId() if callable(sessionId) else sessionId
                if actual_session_id is not None:
                    trace_config.session_id = actual_session_id
                trace = logger.trace(config=trace_config)
                token = _maxim_ctx_var_trace.set(trace)
                _maxim_ctx_var_logger.set(logger)
                try:
                    result = await func(*args, **kwargs)
                    if trace.data().get("output", None) is None:
                        trace.set_output(str(result))
                    return result
                finally:
                    trace.end()
                    _maxim_ctx_var_trace.reset(token)
        else:
            @wraps(func)
            def wrapper(*args, **kwargs):
                actual_trace_id = id() if callable(id) else id
                if actual_trace_id is None:
                    actual_trace_id = str(uuid4())
                trace_config = TraceConfig(id=actual_trace_id, name=name, tags=tags)
                # If session_id is present, we will attach that to the trace
                actual_session_id = sessionId() if callable(sessionId) else sessionId
                if actual_session_id is not None:
                    trace_config.session_id = actual_session_id
                trace = logger.trace(config=trace_config)
                token = _maxim_ctx_var_trace.set(trace)
                _maxim_ctx_var_logger.set(logger)
                try:
                    result = func(*args, **kwargs)
                    if trace.data().get("output", None) is None:
                        trace.set_output(str(result))
                    return result
                finally:
                    trace.end()
                    _maxim_ctx_var_trace.reset(token)
            
        return wrapper

    return decorator



import asyncio
from contextvars import ContextVar
from functools import wraps
from typing import Callable, Dict, Optional
from uuid import uuid4

from ..logger.components.generation import Generation, GenerationConfig
from ..logger.logger import Logger
from .span import current_span
from .trace import current_logger, current_trace

_generation_ctx_var: ContextVar[Optional[Generation]] = ContextVar(
    "maxim_ctx_generation", default=None
)


def current_generation() -> Optional[Generation]:
    return _generation_ctx_var.get()


def generation(
    logger: Optional[Logger] = None,
    id: Optional[Callable] or Optional[str] = None,
    name: Optional[str] = None,
    maxim_prompt_id: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
):
    def decorator(func):
        if asyncio.iscoroutinefunction(func):

            @wraps(func)
            async def wrapper(*args, **kwargs):
                # First check if the logger is available
                maxim_logger = logger
                if maxim_logger is None:
                    if current_logger() is None:
                        raise ValueError(
                            "[MaximSDK]: no logger found. either call this function from a @trace decorated function or pass a logger"
                        )
                    maxim_logger = current_logger()
                if current_span() is None and current_trace() is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                actual_generation_id = id() if callable(id) else id
                if actual_generation_id is None:
                    actual_generation_id = str(uuid4())
                generation_config = GenerationConfig(
                    id=actual_generation_id,
                    name=name,
                    maxim_prompt_id=maxim_prompt_id,
                    tags=tags,
                )
                generation: Generation
                if current_span() is not None:
                    generation = current_span().generation(config=generation_config)
                elif current_trace() is not None:
                    generation = current_trace().generation(config=generation_config)
                token = _generation_ctx_var.set(generation)
                try:
                    return await func(*args, **kwargs)
                finally:
                    generation.end()
                    _generation_ctx_var.reset(token)

        else:

            @wraps(func)
            def wrapper(*args, **kwargs):
                # First check if the logger is available
                maxim_logger = logger
                if maxim_logger is None:
                    if current_logger() is None:
                        raise ValueError(
                            "[MaximSDK]: no logger found. either call this function from a @trace decorated function or pass a logger"
                        )
                    maxim_logger = current_logger()
                if current_span() is None and current_trace() is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                actual_generation_id = id() if callable(id) else id
                if actual_generation_id is None:
                    actual_generation_id = str(uuid4())
                generation_config = GenerationConfig(
                    id=actual_generation_id,
                    name=name,
                    maxim_prompt_id=maxim_prompt_id,
                    tags=tags,
                )
                generation: Generation
                if current_span() is not None:
                    generation = current_span().generation(config=generation_config)
                elif current_trace() is not None:
                    generation = current_trace().generation(config=generation_config)
                token = _generation_ctx_var.set(generation)
                try:
                    return func(*args, **kwargs)
                finally:
                    generation.end()
                    _generation_ctx_var.reset(token)

        return wrapper

    return decorator



from typing import Union


def parse_tags(data: dict) -> Union[None, dict]:
    if data is None:
        return None
    if not isinstance(data, dict):
        raise ValueError("tags must be a dict[str,str]")
    for key, value in data.items():
        if isinstance(value, dict):
            data[key] = parse_tags(value)
        elif not isinstance(value, str):
            try:
                data[key] = str(value)
            except Exception as e:
                data[key] = repr(value)
    return data



from typing import List, Optional

from .models.prompt import RuleGroupType, RuleType


class IncomingQuery():
    def __init__(self, query: str, operator: str, exactMatch: bool):
        self.query = query
        self.operator = operator
        self.exactMatch = exactMatch


class QueryObject():
    def __init__(self, id: str, query: RuleGroupType):
        self.id = id
        self.query = query


def parseIncomingQuery(incomingQuery: str) -> List[RuleType]:
    if not incomingQuery.strip():
        return []
    operators = ["!=", ">=", "<=", ">", "<",
                 "includes", "does not include", "="]
    result = []
    for condition in incomingQuery.split(","):
        noOperatorFound = False
        for op in operators:
            if op in condition:
                no_operator_found = True
                field, value = map(str.strip, condition.split(op))
                exactMatch = False
                if field.startswith("!!"):
                    exactMatch = True
                    field = field[2:]
                # converting 'True'/'False':str to True/False:bool
                if value == 'True' or value == 'False':
                    value = bool(value)
                try:
                    value = int(value)
                except ValueError:
                    pass
                result.append(RuleType(field=field, value=value,
                              operator=op, exactMatch=exactMatch))
        if noOperatorFound:
            raise ValueError(
                f"Unsupported operator found in condition \"{condition}\"")
    return result


def evaluateRuleGroup(ruleGroup: RuleGroupType, incomingQueryRules: List[RuleType]) -> bool:
    matchedRules = []
    matchResults = []
    for rule in ruleGroup.rules:
        if isinstance(rule, RuleGroupType):
            matchResults.append(evaluateRuleGroup(rule, incomingQueryRules))
        else:
            match = False
            for incomingRule in incomingQueryRules:
                if rule.field == incomingRule.field and conditionMet(rule, incomingRule):
                    match = match or True
                    matchedRules.append(incomingRule)
            matchResults.append(match)
            if match:
                matchedRules.append(rule)
    exactMatches = all(
        rule.exactMatch == False or rule in matchedRules for rule in incomingQueryRules)
    if not exactMatches:
        return False
    if ruleGroup.combinator == "AND":
        return all(matchResults)
    else:
        return any(matchResults)

# Had to write this extensive logic in python since type support in python isn't as extensive as typescript


def checkOperatorMatch(fieldRule: RuleType, fieldIncomingRule: RuleType) -> bool:
    operator = fieldRule.operator

    if operator == "=":
        return fieldRule.value == fieldIncomingRule.value

    elif operator == "!=":
        return fieldRule.value != fieldIncomingRule.value

    elif operator == ">":
        if isinstance(fieldRule.value, int) and isinstance(fieldIncomingRule.value, int):
            return fieldRule.value > fieldIncomingRule.value
        elif isinstance(fieldRule.value, str) and isinstance(fieldIncomingRule.value, str):
            return fieldRule.value > fieldIncomingRule.value
        raise ValueError(
            f"Cannot operate {operator} on types {fieldRule} and {fieldIncomingRule}")

    elif operator == "<":
        if isinstance(fieldRule.value, int) and isinstance(fieldIncomingRule.value, int):
            return fieldRule.value < fieldIncomingRule.value
        elif isinstance(fieldRule.value, str) and isinstance(fieldIncomingRule.value, str):
            return fieldRule.value < fieldIncomingRule.value
        raise ValueError(
            f"Cannot operate {operator} on types {fieldRule} and {fieldIncomingRule}")

    elif operator == ">=":
        if isinstance(fieldRule.value, int) and isinstance(fieldIncomingRule.value, int):
            return fieldRule.value >= fieldIncomingRule.value
        elif isinstance(fieldRule.value, str) and isinstance(fieldIncomingRule.value, str):
            return fieldRule.value >= fieldIncomingRule.value

        raise ValueError(
            f"Cannot operate {operator} on types {fieldRule} and {fieldIncomingRule}")

    elif operator == "<=":
        if isinstance(fieldRule.value, int) and isinstance(fieldIncomingRule.value, int):
            return fieldRule.value <= fieldIncomingRule.value
        elif isinstance(fieldRule.value, str) and isinstance(fieldIncomingRule.value, str):
            return fieldRule.value <= fieldIncomingRule.value

        raise ValueError(
            f"Cannot operate {operator} on types {fieldRule} and {fieldIncomingRule}")

    elif operator == "includes":
        if isinstance(fieldRule.value, str) and isinstance(fieldIncomingRule.value, str):
            return fieldIncomingRule.value in fieldRule.value
        elif isinstance(fieldRule.value, list) and isinstance(fieldIncomingRule.value, list):
            return fieldIncomingRule.value in fieldRule.value
        elif isinstance(fieldRule.value, list) and isinstance(fieldIncomingRule.value, str):
            return fieldIncomingRule.value in fieldRule.value
        elif isinstance(fieldRule.value, str) and isinstance(fieldIncomingRule.value, list):
            return fieldRule.value in fieldIncomingRule.value
        raise ValueError(
            f"Cannot operate {operator} on types {fieldRule} and {fieldIncomingRule}")

    elif operator == "does not include":
        if isinstance(fieldRule.value, str) and isinstance(fieldIncomingRule.value, str):
            return fieldIncomingRule.value not in fieldRule.value
        elif isinstance(fieldRule.value, list) and isinstance(fieldIncomingRule.value, list):
            return fieldIncomingRule.value not in fieldRule.value
        elif isinstance(fieldRule.value, list) and isinstance(fieldIncomingRule.value, str):
            return fieldIncomingRule.value not in fieldRule.value
        elif isinstance(fieldRule.value, str) and isinstance(fieldIncomingRule.value, list):
            return fieldRule.value not in fieldIncomingRule.value
        raise ValueError(
            f"Cannot operate {operator} on types {fieldRule} and {fieldIncomingRule}")

    else:
        return False


def conditionMet(fieldRule: RuleType, fieldIncomingRule: RuleType) -> bool:
    if type(fieldRule.value) != type(fieldIncomingRule.value):
        if isinstance(fieldRule.value, int):
            if isinstance(fieldIncomingRule.value, int) or isinstance(fieldIncomingRule.value, bool):
                fieldIncomingRule.value = int(fieldIncomingRule.value)

        # revisit : possible error : everything will be true with bool(fieldIncomingRule.value)
        elif isinstance(fieldRule.value, bool):
            fieldIncomingRule.value = bool(fieldIncomingRule.value)
        elif isinstance(fieldRule.value, str):
            fieldIncomingRule.value = str(fieldIncomingRule.value)

    match = checkOperatorMatch(fieldRule, fieldIncomingRule)
    return match


def findBestMatch(objects: List[QueryObject], incomingQuery: IncomingQuery) -> Optional[QueryObject]:
    bestMatch = None
    maxMatchCount = 0
    incomingQueryRules = parseIncomingQuery(incomingQuery.query)
    for obj in objects:
        if evaluateRuleGroup(obj.query, incomingQueryRules):
            if incomingQuery.exactMatch and len(incomingQueryRules) != len(obj.query.rules):
                continue
            matchCount = len(obj.query.rules)
            if matchCount > maxMatchCount:
                maxMatchCount = matchCount
                bestMatch = obj
    return bestMatch


def findAllMatches(objects: List[QueryObject], incomingQuery: IncomingQuery) -> List[QueryObject]:
    matches: List[QueryObject] = []
    incomingQueryRules = parseIncomingQuery(incomingQuery.query)
    for object in objects:
        if evaluateRuleGroup(object.query, incomingQueryRules):
            if incomingQuery.exactMatch and len(incomingQueryRules) != len(object.query.rules):
                continue
            matches.append(object)
    return matches



from dataclasses import field
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

from ..parsers.tags_parser import parse_tags
from ..writer import LogWriter
from .types import CommitLog, Entity


class ContainerLister:
    def on_end(self):
        pass


BaseConfig = Dict[str, Any]


class BaseContainer:

    def __init__(self, entity: Entity, config: BaseConfig, writer: LogWriter):
        self.entity = entity
        self._id = config['id']
        self._name = config.get('name')
        self.span_id = config.get('span_id')
        self.start_timestamp = datetime.now(timezone.utc)
        self.end_timestamp = None
        self.tags = parse_tags(config.get('tags', {}))
        self.writer = writer

    @property
    def id(self) -> str:
        return self._id

    def add_tag(self, key: str, value: str):
        if self.tags == None:
            self.tags = {}
        if not isinstance(value, str):
            raise ValueError("Tag value must be a string")
        # Validate if value is str and not None
        if not value:
            raise ValueError("Tag value must not be empty")
        self.tags[key] = value
        self.tags = parse_tags(self.tags)
        self._commit("update", {"tags": {key: value}})

    @staticmethod
    def _add_tag_(writer: LogWriter, entity: Entity, id: str, key: str, value: str):
        writer.commit(CommitLog(entity, id, "update", {"tags": {key: value}}))

    def end(self):
        self.end_timestamp = datetime.now(timezone.utc)
        self._commit("end", {"endTimestamp": self.end_timestamp})

    @staticmethod
    def _end_(writer: LogWriter, entity: Entity, id: str, data: Optional[Dict[str, Any]] = None):
        if data is None:
            data = {}
        data = {k: v for k, v in data.items() if v is not None}
        writer.commit(CommitLog(entity, id, "end", data))

    def data(self) -> Dict[str, Any]:
        data = {
            "name": self._name,
            "spanId": self.span_id,
            "tags": self.tags,
            "startTimestamp": self.start_timestamp,
            "endTimestamp": self.end_timestamp,
        }
        # removing none values
        data = {k: v for k, v in data.items() if v is not None}
        return data

    @staticmethod
    def _commit_(writer: LogWriter, entity: Entity, id: str, action: str, data: Optional[Dict[str, Any]] = None):
        # Removing all null values from data dict
        if data is not None:
            data = {k: v for k, v in data.items() if v is not None}
        writer.commit(CommitLog(entity, id, action, data))

    def _commit(self, action: str, data: Optional[Dict[str, Any]] = None):
        if data is None:
            data = self.data()
        # Removing all null values from data dict
        data = {k: v for k, v in data.items() if v is not None}
        self.writer.commit(CommitLog(self.entity, self._id,
                           action, data))


class EventEmittingBaseContainer(BaseContainer):

    @staticmethod
    def _event_(writer: LogWriter, entity: Entity, entity_id: str, id: str, name: str, tags: Optional[Dict[str, str]] = None):
        BaseContainer._commit_(writer, entity, entity_id, "add-event", {
            "id": id, "name": name, "timestamp": datetime.now(timezone.utc), "tags": tags})

    def event(self, id: str, name: str, tags: Optional[Dict[str, str]] = None):
        self._commit("add-event", {"id": id, "name": name,
                                   "timestamp": datetime.now(timezone.utc), "tags": tags})



import json
import logging
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from dataclasses import asdict, dataclass
from typing import Dict, List, Optional, Union

from .apis.maxim import MaximAPI
from .cache.cache import MaximCache
from .cache.inMemory import MaximInMemoryCache
from .filterObjects import (IncomingQuery, QueryObject, findAllMatches,
                            findBestMatch, parseIncomingQuery)
from .logger.logger import Logger, LoggerConfig
from .models.dataset import DatasetEntry
from .models.folder import Folder, FolderEncoder
from .models.prompt import (Prompt, PromptData, PromptVersion, RuleGroupType,
                            RuleType, VersionAndRulesWithPromptId,
                            VersionAndRulesWithPromptIdEncoder,
                            VersionsAndRules, VersionSpecificDeploymentConfig)
from .models.promptChain import (PromptChain, PromptChainRuleGroupType,
                                 PromptChainVersion,
                                 PromptChainVersionsAndRules,
                                 VersionAndRulesWithPromptChainIdEncoder)
from .models.queryBuilder import QueryRule

maximLogger = logging.getLogger("MaximSDK")


@dataclass
class Config():
    api_key: str
    base_url: Optional[str] = None
    cache: Optional[MaximCache] = None
    debug: Optional[bool] = False
    raise_exceptions: Optional[bool] = False
    prompt_management: Optional[bool] = False


EntityType = {
    "PROMPT": "PROMPT",
    "FOLDER": "FOLDER",
    "PROMPT_CHAIN": "PROMPT_CHAIN"
}


class Maxim:
    def __init__(self, config: Config):
        maximLogger.setLevel(logging.DEBUG if config.debug else logging.INFO)
        self.base_url = config.base_url or "https://app.getmaxim.ai"
        self.api_key = config.api_key
        self.is_running = True
        self.raise_exceptions = config.raise_exceptions or False
        self.maxim_api = MaximAPI(self.base_url, self.api_key)
        self.__is_debug = config.debug or False
        self.__loggers: Dict[str, Logger] = {}
        self.prompt_management = config.prompt_management
        if self.prompt_management:
            self.__cache = config.cache or MaximInMemoryCache()
            self.__sync_thread = threading.Thread(target=self.__sync_timer)
            self.__sync_thread.daemon = True
            self.__sync_thread.start()

    def __sync_timer(self):
        if not self.prompt_management:
            return
        while self.is_running:
            self.__sync_entities()
            time.sleep(60)

    def __sync_entities(self):
        maximLogger.debug(" Syncing prompts and folders")
        if not self.prompt_management:
            return
        if not self.is_running:
            return
        self.__sync_prompts()
        self.__sync_prompt_chains()
        self.__syncFolders()
        maximLogger.debug("Syncing completed")

    def __sync_prompts(self):
        maximLogger.debug("Syncing prompts")
        try:
            prompts = self.maxim_api.get_prompts()
            maximLogger.debug(f"Found {len(prompts)} prompts")
            for prompt in prompts:
                try:
                    self.__cache.set(
                        self.__get_cache_key(EntityType["PROMPT"], prompt.promptId),
                        json.dumps(prompt, cls=VersionAndRulesWithPromptIdEncoder),
                    )
                except Exception as err:
                    maximLogger.error(err)
                    if self.raise_exceptions:
                        raise err
        except Exception as err:
            maximLogger.error(f"Error while syncing prompts: {err}")
            if self.raise_exceptions:
                raise err

    def __sync_prompt_chains(self):
        maximLogger.debug("Syncing prompt Chains")
        try:
            promptChains = self.maxim_api.get_prompt_chains()
            maximLogger.debug(f"Found {len(promptChains)} prompt chains")
            for promptChain in promptChains:
                try:
                    self.__cache.set(
                        self.__get_cache_key(
                            EntityType["PROMPT_CHAIN"], promptChain.promptChainId
                        ),
                        json.dumps(
                            promptChain, cls=VersionAndRulesWithPromptChainIdEncoder
                        ),
                    )
                except Exception as err:
                    maximLogger.error(err)
                    if self.raise_exceptions:
                        raise err
        except Exception as err:
            maximLogger.error(f"Error while syncing prompts: {err}")
            if self.raise_exceptions:
                raise err

    def __syncFolders(self):
        maximLogger.debug("Syncing folders")
        try:
            folders = self.maxim_api.get_folders()
            maximLogger.debug(f"Found {len(folders)} folders")
            for folder in folders:
                try:
                    self.__cache.set(
                        self.__get_cache_key(EntityType["FOLDER"], folder.id),
                        json.dumps(folder, cls=FolderEncoder),
                    )
                except Exception as err:
                    maximLogger.error(err)
                    if self.raise_exceptions:
                        raise err
        except Exception as err:
            maximLogger.error(f"Error while syncing folders: {err}")
            if self.raise_exceptions:
                raise err

    def __get_cache_key(self, entity: str, id: str) -> str:
        if entity == 'PROMPT':
            return f"prompt:{id}"
        else:
            return f"folder:{id}"

    def __get_prompt_from_cache(self, key: str) -> Optional[VersionsAndRules]:
        data = self.__cache.get(key)
        if not data:
            return None
        json_data = json.loads(data)
        return VersionAndRulesWithPromptId.from_dict(json_data)

    def __get_all_prompts_from_cache(self) -> Optional[List[VersionsAndRules]]:
        keys = self.__cache.get_all_keys()
        if not keys:
            return None
        data = [self.__cache.get(key)
                for key in keys if key.startswith("prompt:")]
        promptList = []
        for d in data:
            if d is not None:
                json_data = json.loads(d)
                if 'promptId' in json_data:
                    json_data.pop('promptId')
                promptList.append(VersionsAndRules.from_dict(json_data))
        return promptList

    def __get_prompt_chain_from_cache(
        self, key: str
    ) -> Optional[PromptChainVersionsAndRules]:
        data = self.__cache.get(key)
        if not data:
            return None
        parsed_data = json.loads(data)
        return PromptChainVersionsAndRules.from_dict(parsed_data)

    def __get_all_prompt_chains_from_cache(
        self,
    ) -> Optional[List[PromptChainVersionsAndRules]]:
        keys = self.__cache.get_all_keys()
        if not keys:
            return None
        data = [self.__cache.get(key)
                for key in keys if key.startswith("prompt_chain:")]
        promptChainList = []
        for d in data:
            if d is not None:
                json_data = json.loads(d)
                if 'promptChainId' in json_data:
                    json_data.pop('promptChainId')
                promptChainList.append(
                    PromptChainVersionsAndRules.from_dict(json_data))
        return promptChainList

    def __get_folder_from_cache(self, key: str) -> Optional[Folder]:
        data = self.__cache.get(key)
        if not data:
            return None
        json_data = json.loads(data)
        return Folder(**json_data)

    def __get_all_folders_from_cache(self) -> Optional[List[Folder]]:
        keys = self.__cache.get_all_keys()
        if not keys:
            return None
        data = [self.__cache.get(key)
                for key in keys if key.startswith("folder:")]
        return [Folder(**json.loads(d)) for d in data if d is not None]

    def __format_prompt(self, prompt: PromptVersion) -> Prompt:
        json_data = {
            'promptId': prompt.promptId,
            'version': prompt.version,
            'versionId': prompt.id,
            'tags': prompt.config.tags if prompt.config else [],
            'model': prompt.config.model if prompt.config else None,
            'messages': prompt.config.messages if prompt.config else None,
            'modelParameters': prompt.config.modelParameters if prompt.config else None
        }
        return Prompt(**json_data)

    def __format_prompt_chain(
        self, promptChainVersion: PromptChainVersion
    ) -> PromptChain:
        json_data = {
            'promptChainId': promptChainVersion.promptChainId,
            'versionId': promptChainVersion.id,
            'versionId': promptChainVersion.version,
            'nodes': [n for n in promptChainVersion.config.nodes if hasattr(n, 'prompt') or (isinstance(n, dict) and 'prompt' in n)] if promptChainVersion.config else []
        }
        return PromptChain(**json_data)

    def __get_prompt_version_for_rule(
        self,
        promptVersionAndRules: VersionAndRulesWithPromptId,
        rule: Optional[QueryRule] = None,
    ) -> Optional[Prompt]:
        if rule:
            incomingQuery = IncomingQuery(
                query=rule.query, operator=rule.operator, exactMatch=rule.exactMatch)
            objects = []
            for versionId, versionRules in promptVersionAndRules.rules.items():
                for versionRule in versionRules:
                    if not versionRule.rules.query:
                        continue
                    if rule.scopes:
                        for key in rule.scopes.keys():
                            if key != "folder":
                                raise ValueError("Invalid scope added")
                    version = next(
                        (v for v in promptVersionAndRules.versions if v.id == versionId), None)
                    if not version:
                        continue
                    query = versionRule.rules.query
                    if version.config and version.config.tags:
                        parsedIncomingQuery = parseIncomingQuery(
                            incomingQuery.query)
                        tags = version.config.tags
                        for key, value in tags.items():
                            if value == None:
                                continue
                            if parsedIncomingQuery is None:
                                continue
                            incomingQueryFields = [
                                q.field for q in parsedIncomingQuery]
                            if key in incomingQueryFields:
                                query.rules.append(
                                    RuleType(field=key, operator='=', value=value))
                    objects.append(QueryObject(query=query, id=versionId))

            deployedVersionObject = findBestMatch(objects, incomingQuery)
            if deployedVersionObject:
                deployedVersion = next(
                    (v for v in promptVersionAndRules.versions if v.id == deployedVersionObject.id), None)
                if deployedVersion:
                    return self.__format_prompt(deployedVersion)

        else:
            if promptVersionAndRules.rules:
                for versionId, versionRules in promptVersionAndRules.rules.items():
                    logging.debug(f"type: {type(versionRules[0].rules)}")
                    doesQueryExist = any(
                        ruleElm.rules.query != None for ruleElm in versionRules)
                    if doesQueryExist:
                        deployedVersion = next(
                            (v for v in promptVersionAndRules.versions if v.id == versionId), None)
                        if deployedVersion:
                            return self.__format_prompt(deployedVersion)
            else:
                return self.__format_prompt(promptVersionAndRules.versions[0])

        if promptVersionAndRules.fallbackVersion:
            maximLogger.info(
                f"************* FALLBACK TRIGGERED : VERSION : {promptVersionAndRules.fallbackVersion} *************")
            return self.__format_prompt(promptVersionAndRules.fallbackVersion)
        return None

    def __get_prompt_chain_version_for_rule(
        self,
        promptChainVersionsAndRules: PromptChainVersionsAndRules,
        rule: Optional[QueryRule] = None,
    ) -> Optional[PromptChain]:
        if rule:
            incomingQuery = IncomingQuery(
                query=rule.query, operator=rule.operator, exactMatch=rule.exactMatch)
            objects = []
            for versionId, versionRules in promptChainVersionsAndRules.rules.items():
                for versionRule in versionRules:
                    if not versionRule.rules.query:
                        continue
                    if rule.scopes:
                        for key in rule.scopes.keys():
                            if key != "folder":
                                raise ValueError("Invalid scope added")
                    version = next(
                        (v for v in promptChainVersionsAndRules.versions if v.id == versionId), None)
                    if not version:
                        continue
                    query = RuleGroupType(**versionRule.rules.query.__dict__)
                    objects.append(QueryObject(query=query, id=versionId))

            deployedVersionObject = findBestMatch(objects, incomingQuery)
            if deployedVersionObject:
                deployedVersion = next(
                    (v for v in promptChainVersionsAndRules.versions if v.id == deployedVersionObject.id), None)
                if deployedVersion:
                    return self.__format_prompt_chain(deployedVersion)

        else:
            if promptChainVersionsAndRules.rules:
                for versionId, versionRules in promptChainVersionsAndRules.rules.items():
                    logging.debug(f"type: {type(versionRules[0].rules)}")
                    doesQueryExist = any(
                        ruleElm.rules.query != None for ruleElm in versionRules)
                    if doesQueryExist:
                        deployedVersion = next(
                            (v for v in promptChainVersionsAndRules.versions if v.id == versionId), None)
                        if deployedVersion:
                            return self.__format_prompt_chain(deployedVersion)
            else:
                return self.__format_prompt_chain(
                    promptChainVersionsAndRules.versions[0]
                )

        if promptChainVersionsAndRules.fallbackVersion:
            maximLogger.info(
                f"************* FALLBACK TRIGGERED : VERSION : {promptChainVersionsAndRules.fallbackVersion} *************")
            return self.__format_prompt_chain(
                promptChainVersionsAndRules.fallbackVersion
            )
        return None

    def __get_folders_for_rule(
        self, folders: List[Folder], rule: QueryRule
    ) -> List[Folder]:
        incomingQuery = IncomingQuery(
            query=rule.query, operator=rule.operator, exactMatch=rule.exactMatch)
        objects = []
        for folder in folders:
            query = RuleGroupType(rules=[], combinator='AND')
            if not folder.tags:
                continue
            parsedIncomingQuery = parseIncomingQuery(incomingQuery.query)
            tags = folder.tags
            for key, value in tags.items():
                if key in [q.field for q in parsedIncomingQuery]:
                    # if not isinstance(value,None):
                    query.rules.append(
                        RuleType(field=key, operator='=', value=value))
            if not query.rules:
                continue
            objects.append(QueryObject(query=query, id=folder.id))

        folderObjects = findAllMatches(objects, incomingQuery)
        ids = [fo.id for fo in folderObjects]
        return [f for f in folders if f.id in ids]

    def get_prompt(self, id: str, rule: QueryRule) -> Optional[Prompt]:
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config."
            )
        maximLogger.debug(f"getPrompt: id: {id} for rule: {rule}")
        key = self.__get_cache_key("PROMPT", id)
        versionAndRulesWithPromptId = self.__get_prompt_from_cache(key)
        if versionAndRulesWithPromptId is None:
            versionAndRulesWithPromptId = self.maxim_api.get_prompt(id)
            if len(versionAndRulesWithPromptId.versions) == 0:
                return None
            self.__cache.set(id, json.dumps(
                versionAndRulesWithPromptId, cls=VersionAndRulesWithPromptIdEncoder))
        if not versionAndRulesWithPromptId:
            return None
        prompt = self.__get_prompt_version_for_rule(versionAndRulesWithPromptId, rule)
        if not prompt:
            return None
        return prompt

    def get_prompts(self, rule: QueryRule) -> List[Prompt]:
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config."
            )
        versionAndRules = self.__get_all_prompts_from_cache()
        prompts = []
        if versionAndRules is None or len(versionAndRules) == 0:
            self.__sync_entities()
            versionAndRules = self.__get_all_prompts_from_cache()
        if not versionAndRules:
            return []
        for v in versionAndRules:
            if rule.scopes:
                if 'folder' not in rule.scopes.keys():
                    return []
                else:
                    if rule.scopes['folder'] != v.folderId:
                        continue
            prompt = self.__get_prompt_version_for_rule(v, rule)
            if prompt != None:
                prompts.append(prompt)
        if len(prompts) == 0:
            return []
        return prompts

    def get_prompt_chain(self, id: str, rule: QueryRule) -> Optional[PromptChain]:
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config."
            )
        key = self.__get_cache_key("PROMPT_CHAIN", id)
        versionAndRules = self.__get_prompt_chain_from_cache(key)
        if versionAndRules is None:
            versionAndRules = self.maxim_api.getPromptChain(id)
            if len(versionAndRules.versions) == 0:
                return None
            self.__cache.set(id, json.dumps(
                versionAndRules, cls=VersionAndRulesWithPromptChainIdEncoder))
        if not versionAndRules:
            return None
        promptChains = self.__get_prompt_chain_version_for_rule(versionAndRules, rule)
        if not promptChains:
            return None
        return promptChains

    def get_folder_by_id(self, id: str) -> Optional[Folder]:
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config."
            )
        key = self.__get_cache_key("FOLDER", id)
        folder = self.__get_folder_from_cache(key)
        if folder is None:
            try:
                folder = self.maxim_api.get_folder(id)
                if not folder:
                    return None
                self.__cache.set(key, json.dumps(folder, cls=FolderEncoder))
            except Exception as e:
                return None
        return folder

    def get_folders(self, rule: QueryRule) -> List[Folder]:
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config."
            )
        folders = self.__get_all_folders_from_cache()
        if folders is None or len(folders) == 0:
            self.__sync_entities()
            folders = self.__get_all_folders_from_cache()
        if not folders:
            return []
        return self.__get_folders_for_rule(folders, rule)

    def logger(self, config: LoggerConfig) -> Logger:
        # Checking if this log repository exist on server
        exists = self.maxim_api.does_log_repository_exist(config.id)
        if not exists:
            if config.id:
                maximLogger.error(f"Log repository not found")
                if self.raise_exceptions:
                    raise Exception(f"Log repository not found")
        if config.id in self.__loggers:
            return self.__loggers[config.id]
        logger = Logger(
            config=config,
            api_key=self.api_key,
            base_url=self.base_url,
            is_debug=self.__is_debug,
            raise_exceptions=self.raise_exceptions,
        )
        self.__loggers[config.id] = logger
        return logger

    def cleanup(self):
        maximLogger.debug("Cleaning up Maxim sync thread")
        self.is_running = False
        for logger in self.__loggers.values():
            logger.cleanup()
        maximLogger.debug("Cleanup done")
        logger.cleanup()
        maximLogger.debug("Cleanup done")



import json
from typing import Any, List
from uuid import uuid4

from .core import (validate_content, validate_optional_type, validate_type,
                   validate_type_to_be_one_of)


def parse_function_call(function_call_data):
    validate_type(function_call_data.get('name'), str, 'name')
    validate_type(function_call_data.get('arguments'), str, 'arguments')
    return function_call_data


def parse_tool_calls(tool_calls_data):
    validate_type(tool_calls_data.get('id'), str, 'id')
    validate_type(tool_calls_data.get('type'), str, 'type')
    parse_function_call(tool_calls_data.get('function'))
    return tool_calls_data


def parse_chat_completion_choice(messages_data):
    validate_type(messages_data.get('role'), str, 'role')
    validate_optional_type(messages_data.get('content'), str, 'content')
    if messages_data.get('function_call') is not None:
        parse_function_call(messages_data.get('function_call'))
    elif messages_data.get('tool_calls') is not None:
        parse_tool_calls(messages_data.get('tool_calls'))
    return messages_data


def parse_choice(choice_data):
    validate_type(choice_data.get('index'), int, 'index')
    validate_type(choice_data.get('finish_reason'), str, 'finish_reason')

    # Checking if text completion or chat completion
    if choice_data.get('text') is not None:
        validate_type(choice_data.get('text'), str, 'text')
    elif choice_data.get('messages') is not None:
        parse_chat_completion_choice(choice_data.get('messages'))

    return choice_data


def parse_usage(usage_data):
    if usage_data is None:
        return None
    validate_type(usage_data.get('prompt_tokens'), int, 'prompt_tokens')
    validate_type(usage_data.get('completion_tokens'),
                  int, 'completion_tokens')
    validate_type(usage_data.get('total_tokens'), int, 'total_tokens')
    return usage_data


def parse_generation_error(error_data):
    if error_data is None:
        return None
    validate_type(error_data.get('message'), str, 'message')
    validate_optional_type(error_data.get('code'), str, 'code')
    validate_optional_type(error_data.get('type'), str, 'type')
    return error_data


def parse_result(data: dict) -> dict:
    if not isinstance(data, dict):
        raise ValueError("Text completion is not supported.")
    validate_type(data.get('id'), str, 'id')
    validate_optional_type(data.get('object'), str, 'object')
    validate_type(data.get('created'), int, 'created')
    validate_optional_type(data.get('model'), str, 'model')

    choices_data = data.get('choices')
    validate_type(choices_data, list, 'choices')
    if not choices_data:
        raise ValueError("choices must not be empty")
    choices = [parse_choice(choice) for choice in choices_data]
    usage = parse_usage(data.get('usage', None))
    error = parse_generation_error(data.get('error', None))
    result = {
        'id': data['id'],
        'object': data['object'] if 'object' in data else None,
        'created': data['created'],
        'choices': choices,
        'usage': usage,
        'error': error if error else None
    }
    # removing all None from result
    result = {k: v for k, v in result.items() if v is not None}
    return result


def parse_message(message: Any) -> Any:
    validate_type(message.get('role'), str, 'role')
    validate_content(message.get('role'), [
                     'user', 'assistant', 'system', 'bot', 'chatbot', 'model'])
    validate_type_to_be_one_of(message.get('content'), [str, object], 'type')
    if isinstance(message.get('content'), object):
        # Making sure if content has type and corresponding data
        content = message.get('content')
        validate_type(content.get('type'), str, 'type')
        validate_content(content.get('type'), ['image_url', 'text'])
        # Making sure type is image or text
        type = content.get('type')
        if type == 'image_url':
            validate_type(content.get('image_url'), str, 'image_url')
        elif type == 'text':
            validate_type(content.get('text'), str, 'text')
        else:
            raise ValueError(
                f"Invalid content type. We expect 'text' or 'image' type. Got: {type}")
    return message


def parse_messages(messages: List[Any]) -> List[Any]:
    if len(messages) == 0:
        return []
    return [parse_message(message) for message in messages]


def parse_model_parameters(parameters: dict) -> dict:
    # convert parameters dict into JSON string
    if parameters is None:
        return {}
    # we will go through each key and make sure it is a string
    # if not we will do json.dumps on it
    for key, value in parameters.items():
        if not isinstance(value, str):
            parameters[key] = json.dumps(value)
    return parameters



import asyncio
import logging
from contextvars import ContextVar
from functools import wraps
from typing import Any, Dict, Optional

from ...logger.langchain import MaximLangchainTracer
from ...logger.logger import Logger
from ..span import current_span
from ..trace import current_logger, current_trace

_langchain_tracer_ctx_var: ContextVar[Optional[MaximLangchainTracer]] = ContextVar(
    "maxim_ctx_langchain_tracer", default=None
)


def langchain_callback() -> Optional[MaximLangchainTracer]:
    return _langchain_tracer_ctx_var.get(None)


async def async_core(func, logger, name, tags, *args, **kwargs):
    # First check if the logger is available
    maxim_logger = logger
    if maxim_logger is None:
        if current_logger() is None:
            raise ValueError(
                "[MaximSDK]: no logger found. either call this function from a @trace decorated function or pass a logger"
            )
        maxim_logger = current_logger()
        if not isinstance(maxim_logger, Logger):
            raise TypeError("[MaximSDK]: logger must be an instance of Logger")
    # Here there should be an active span or active trace
    # If none of this is the case then we raise an error
    if current_span() is None and current_trace() is None:
        if maxim_logger.raise_exceptions:
            raise ValueError(
                "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
            )
        else:
            logging.warning(
                "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
            )
    # This is a valid call
    metadata: Dict[str, Any] = {}
    if current_span() is not None:
        metadata["span_id"] = current_span().id
    elif current_trace() is not None:
        metadata["trace_id"] = current_trace().id
    if tags is not None:
        metadata["generation_tags"] = tags
    if name is not None:
        metadata["generation_name"] = name
    existing_tracer = _langchain_tracer_ctx_var.get(None)
    if not existing_tracer:
        tracer = MaximLangchainTracer(logger=maxim_logger, metadata=metadata)
        _langchain_tracer_ctx_var.set(tracer)
    return await func(*args, **kwargs)

def sync_core(func, logger, name, tags, *args, **kwargs):
    # First check if the logger is available
    maxim_logger = logger
    if maxim_logger is None:
        if current_logger() is None:
            raise ValueError(
                "[MaximSDK]: no logger found. either call this function from a @trace decorated function or pass a logger"
            )
        maxim_logger = current_logger()
        if not isinstance(maxim_logger, Logger):
            raise TypeError("[MaximSDK]: logger must be an instance of Logger")
    # Here there should be an active span or active trace
    # If none of this is the case then we raise an error
    if current_span() is None and current_trace() is None:
        if maxim_logger.raise_exceptions:
            raise ValueError(
                "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
            )
        else:
            logging.warning(
                "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
            )
    # This is a valid call
    metadata: Dict[str, Any] = {}
    if current_span() is not None:
        metadata["span_id"] = current_span().id
    elif current_trace() is not None:
        metadata["trace_id"] = current_trace().id
    if tags is not None:
        metadata["generation_tags"] = tags
    if name is not None:
        metadata["generation_name"] = name
    existing_tracer = _langchain_tracer_ctx_var.get(None)
    if not existing_tracer:
        tracer = MaximLangchainTracer(logger=maxim_logger, metadata=metadata)
        _langchain_tracer_ctx_var.set(tracer)
    return func(*args, **kwargs)



def langchain_llm_call(
    logger: Optional[Logger] = None,
    name: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
):
    def decorator(func):
        if asyncio.iscoroutinefunction(func):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                return await async_core(func, logger, name, tags, *args, **kwargs)
        else:
            @wraps(func)
            def wrapper(*args, **kwargs):
                return sync_core(func, logger, name, tags, *args, **kwargs)

        return wrapper

    return decorator


def langgraph_agent(
    logger: Optional[Logger] = None,
    name: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
):
    def decorator(func):
        if asyncio.iscoroutinefunction(func):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                return await async_core(func, logger, name, tags, *args, **kwargs)
        else:
            @wraps(func)
            def wrapper(*args, **kwargs):
                return sync_core(func, logger, name, tags, *args, **kwargs)
        return wrapper

    return decorator



from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Dict, List, Optional

from ..writer import LogWriter
from .base import EventEmittingBaseContainer
from .generation import Generation, GenerationConfig
from .retrieval import Retrieval, RetrievalConfig
from .toolCall import ToolCall, ToolCallConfig
from .trace import Trace
from .types import Entity


@dataclass
class SpanConfig:
    id: str
    name: Optional[str] = None
    tags: Optional[Dict[str, str]] = None


class Span(EventEmittingBaseContainer):
    ENTITY = Entity.SPAN

    def __init__(self, config: SpanConfig, writer: LogWriter):
        super().__init__(self.ENTITY, config.__dict__, writer)
        self.traces: List[Trace] = []

    def span(self, config: SpanConfig):
        span = Span(config, self.writer)
        span.span_id = self.id
        self._commit(
            "add-span",
            {
                "id": config.id,
                **span.data(),
            },
        )
        return span

    def input(self, input: str):
        self._commit("update", {input: {type: "text", value: input}})

    @staticmethod
    def input_(writer: LogWriter, span_id: str, input: str):
        Span._commit_(
            writer,
            Entity.SPAN,
            span_id,
            "update",
            {"input": {type: "text", value: input}},
        )

    def output(self, output: str):
        self._commit("update", {output: {type: "text", value: output}})

    @staticmethod
    def output_(writer: LogWriter, span_id: str, output: str):
        Span._commit_(
            writer,
            Entity.SPAN,
            span_id,
            "update",
            {"output": {type: "text", value: output}},
        )

    @staticmethod
    def span_(writer: LogWriter, span_id: str, config: SpanConfig):
        span = Span(config, writer)
        span.span_id = span_id
        Span._commit_(
            writer,
            Entity.SPAN,
            span_id,
            "add-span",
            {
                "id": config.id,
                **span.data(),
            },
        )
        return span

    def generation(self, config: GenerationConfig) -> Generation:
        generation = Generation(config, self.writer)
        payload = generation.data()
        payload["id"] = config.id
        payload["spanId"] = self.id
        self._commit(
            "add-generation",
            {
                **payload,
            },
        )
        return generation

    def tool_call(self, config: ToolCallConfig) -> ToolCall:
        tool_call = ToolCall(config, self.writer)
        self._commit(
            "add-tool-call",
            {
                **tool_call.data(),
                "id": tool_call.id,
            },
        )
        return tool_call

    @staticmethod
    def tool_call_(writer: LogWriter, span_id: str, config: ToolCallConfig) -> ToolCall:
        tool_call = ToolCall(config, writer)
        Span._commit_(
            writer,
            Entity.SPAN,
            span_id,
            "add-tool-call",
            {
                **tool_call.data(),
                "id": tool_call.id,
            },
        )
        return tool_call

    @staticmethod
    def generation_(
        writer: LogWriter, span_id: str, config: GenerationConfig
    ) -> Generation:
        generation = Generation(config, writer)
        Span._commit_(
            writer,
            Entity.SPAN,
            span_id,
            "add-generation",
            {
                **generation.data(),
                "id": config.id,
            },
        )
        return generation

    def retrieval(self, config: RetrievalConfig):
        retrieval = Retrieval(config, self.writer)
        self._commit(
            "add-retrieval",
            {
                "id": config.id,
                **retrieval.data(),
            },
        )
        return retrieval

    @staticmethod
    def retrieval_(writer: LogWriter, span_id: str, config: RetrievalConfig):
        retrieval = Retrieval(config, writer)
        Span._commit_(
            writer,
            Entity.SPAN,
            span_id,
            "add-retrieval",
            {
                "id": config.id,
                **retrieval.data(),
            },
        )
        return retrieval

    @staticmethod
    def end_(writer: LogWriter, span_id: str, data: Optional[Dict[str, str]] = None):
        if data is None:
            data = {}
        return EventEmittingBaseContainer._end_(
            writer,
            Entity.SPAN,
            span_id,
            {"endTimestamp": datetime.now(timezone.utc), **data},
        )

    @staticmethod
    def add_tag_(writer: LogWriter, span_id: str, key: str, value: str):
        return EventEmittingBaseContainer._add_tag_(
            writer, Entity.SPAN, span_id, key, value
        )

    @staticmethod
    def event_(
        writer: LogWriter,
        span_id: str,
        id: str,
        name: str,
        data: Optional[Dict[str, str]],
    ):
        return EventEmittingBaseContainer._event_(
            writer, Entity.SPAN, span_id, id, name, data
        )



# Assuming similar components exist in Python with the same functionality
import logging
from typing import Any, Dict, Optional

from typing_extensions import deprecated

from ..logger.components.session import Session, SessionConfig
from ..logger.components.trace import Trace, TraceConfig
from .components.feedback import Feedback
from .components.generation import (Generation, GenerationConfig,
                                    GenerationError)
from .components.retrieval import Retrieval, RetrievalConfig
from .components.span import Span, SpanConfig
from .components.toolCall import ToolCall, ToolCallConfig, ToolCallError
from .writer import LogWriter, LogWriterConfig


class LoggerConfig:
    def __init__(self, id, auto_flush=True, flush_interval=10):
        self.id = id
        self.auto_flush = auto_flush
        self.flush_interval = flush_interval


logger = logging.getLogger("MaximSDK")


class Logger:

    def __init__(
        self,
        config: LoggerConfig,
        api_key,
        base_url,
        is_debug=False,
        raise_exceptions=False,
    ):
        logger.setLevel(logging.DEBUG if is_debug else logging.INFO)
        if not config.id:
            raise ValueError("Logger must be initialized with id of the logger")
        self._id = config.id
        self.raise_exceptions = raise_exceptions
        self.is_debug = is_debug
        writer_config = LogWriterConfig(
            auto_flush=config.auto_flush,
            flush_interval=config.flush_interval,
            base_url=base_url,
            api_key=api_key,
            is_debug=is_debug,
            repository_id=config.id,
        )
        self.writer = LogWriter(writer_config)
        logger.debug(f"Logger initialized")

    def session(self, config: SessionConfig) -> Session:
        return Session(config, self.writer)

    def trace(self, config: TraceConfig) -> Trace:
        return Trace(config, self.writer)

    # Session methods
    def session_add_tag(self, session_id: str, key: str, value: str):
        Session.add_tag_(self.writer, session_id, key, value)

    def session_end(self, session_id: str):
        Session.end_(self.writer, session_id)

    def session_event(self, session_id: str, event_id: str, event: str, data: Any):
        Session.event_(self.writer, session_id, event_id, event, data)

    def session_feedback(self, session_id: str, feedback: Feedback):
        Session.feedback_(self.writer, session_id, feedback)

    def session_trace(self, session_id: str, config: TraceConfig) -> Trace:
        return Session.trace_(self.writer, session_id, config)

    # Trace methods
    @deprecated(
        "This method will be removed in a future version. Use trace_add_generation instead."
    )
    def trace_generation(self, trace_id: str, config: GenerationConfig) -> Generation:
        return Trace.generation_(self.writer, trace_id, config)

    def trace_add_generation(
        self, trace_id: str, config: GenerationConfig
    ) -> Generation:
        return Trace.generation_(self.writer, trace_id, config)

    @deprecated(
        "This method will be removed in a future version. Use trace_add_retrieval instead."
    )
    def trace_retrieval(self, trace_id: str, config: RetrievalConfig) -> Retrieval:
        return Trace.retrieval_(self.writer, trace_id, config)

    def trace_add_retrieval(self, trace_id: str, config: RetrievalConfig) -> Retrieval:
        return Trace.retrieval_(self.writer, trace_id, config)

    @deprecated(
        "This method will be removed in a future version. Use trace_add_span instead."
    )
    def trace_span(self, trace_id: str, config: SpanConfig) -> Span:
        return Trace.span_(self.writer, trace_id, config)

    def trace_add_span(self, trace_id: str, config: SpanConfig) -> Span:
        return Trace.span_(self.writer, trace_id, config)

    def trace_add_tag(self, trace_id: str, key: str, value: str):
        Trace.add_tag_(self.writer, trace_id, key, value)

    def trace_add_tool_call(self, trace_id: str, config: ToolCallConfig) -> ToolCall:
        return Trace.tool_call_(self.writer, trace_id, config)

    def trace_event(
        self,
        trace_id: str,
        event_id: str,
        event: str,
        tags: Optional[Dict[str, str]] = None,
    ):
        Trace.event_(self.writer, trace_id, event_id, event, tags)

    def trace_set_input(self, trace_id: str, input: str):
        Trace.set_input_(self.writer, trace_id, input)

    def trace_set_output(self, trace_id: str, output: str):
        Trace.set_output_(self.writer, trace_id, output)

    def trace_feedback(self, trace_id: str, feedback: Feedback):
        Trace.feedback_(self.writer, trace_id, feedback)

    def trace_end(self, trace_id: str):
        Trace.end_(self.writer, trace_id)

    # Generation methods
    def generation_set_model(self, generation_id: str, model: str):
        Generation.set_model_(self.writer, generation_id, model)

    def generation_add_message(self, generation_id: str, message: Any):
        Generation.add_message_(self.writer, generation_id, message)

    def generation_set_model_parameters(
        self, generation_id: str, model_parameters: dict
    ):
        Generation.set_model_parameters_(self.writer, generation_id, model_parameters)

    def generation_result(self, generation_id: str, result: Any):
        Generation.result_(self.writer, generation_id, result)

    def generation_end(self, generation_id: str):
        Generation.end_(self.writer, generation_id)

    def generation_error(self, generation_id: str, error: GenerationError):
        Generation.error_(self.writer, generation_id, error)

    # Span methods
    @deprecated(
        "This method will be removed in a future version. Use span_add_generation instead."
    )
    def span_generation(self, span_id: str, config: GenerationConfig) -> Generation:
        return Span.generation_(self.writer, span_id, config)

    def span_add_generation(self, span_id: str, config: GenerationConfig) -> Generation:
        return Span.generation_(self.writer, span_id, config)

    @deprecated(
        "This method will be removed in a future version. Use span_add_retrieval instead."
    )
    def span_retrieval(self, span_id: str, config: RetrievalConfig):
        return Span.retrieval_(self.writer, span_id, config)

    def span_add_retrieval(self, span_id: str, config: RetrievalConfig):
        return Span.retrieval_(self.writer, span_id, config)

    def span_add_tool_call(self, span_id: str, config: ToolCallConfig) -> ToolCall:
        return Span.tool_call_(self.writer, span_id, config)

    def span_end(self, span_id: str):
        Span.end_(self.writer, span_id)

    def span_add_tag(self, span_id: str, key: str, value: str):
        Span.add_tag_(self.writer, span_id, key, value)

    def span_event(
        self,
        span_id: str,
        event_id: str,
        name: str,
        tags: Optional[Dict[str, str]] = None,
    ):
        Span.event_(self.writer, span_id, event_id, name, tags)

    @deprecated(
        "This method will be removed in a future version. Use span_add_sub_span instead."
    )
    def span_span(self, span_id: str, config: SpanConfig):
        return Span.span_(self.writer, span_id, config)

    def span_add_sub_span(self, span_id: str, config: SpanConfig):
        return Span.span_(self.writer, span_id, config)

    # Retrieval methods
    def retrieval_end(self, retrieval_id: str):
        Retrieval.end_(self.writer, retrieval_id)

    def retrieval_input(self, retrieval_id: str, query: Any):
        Retrieval.input_(self.writer, retrieval_id, query)

    def retrieval_output(self, retrieval_id: str, docs: Any):
        Retrieval.output_(self.writer, retrieval_id, docs)

    # Tool call methods

    def tool_call_result(self, tool_call_id: str, result: Any):
        ToolCall.result_(self.writer, tool_call_id, result)

    def tool_call_error(self, tool_call_id: str, error: ToolCallError):
        ToolCall.error_(self.writer, tool_call_id, error)

    @property
    def id(self):
        return self._id

    def flush(self):
        self.writer.flush()

    def cleanup(self):
        self.writer.cleanup()



from dataclasses import asdict, dataclass
import json
from typing import List, Optional, Union, Dict
from datetime import datetime
from .prompt import Prompt

@dataclass
class PromptNode:
    prompt: Prompt

@dataclass
class CodeBlockNode:
    code: str

@dataclass
class ApiParams:
    id: str
    key: str
    value: str

@dataclass
class ApiNode:
    api: dict

@dataclass
class Node:
    order: int
    content: Union[PromptNode, CodeBlockNode, ApiNode]

    @staticmethod
    def from_dict(data: Dict) -> 'Node':
        content_type = next(iter(set(data.keys()) - {'order'}))
        content_class = {
            'prompt': PromptNode,
            'code': CodeBlockNode,
            'api': ApiNode
        }[content_type]
        content_data = data[content_type]
        if content_type == 'prompt':
            content = PromptNode(prompt=Prompt.from_dict(content_data))
        elif content_type == 'code':
            content = CodeBlockNode(code=content_data)
        else:  # api
            content = ApiNode(api=content_data)
        return Node(order=data['order'], content=content)

@dataclass
class PromptChain:
    promptChainId: str
    version: int
    versionId: str
    nodes: List[Node]

@dataclass
class PromptChainVersionConfig:
    nodes: List[Node]

    @staticmethod
    def from_dict(data: Dict) -> 'PromptChainVersionConfig':
        return PromptChainVersionConfig(
            nodes=[Node.from_dict(node) for node in data['nodes']]
        )

@dataclass
class PromptChainVersion:
    id: str
    version: int
    promptChainId: str
    description: Optional[str]
    config: Optional[PromptChainVersionConfig]
    createdAt: str
    updatedAt: str

    @staticmethod
    def from_dict(data: Dict) -> 'PromptChainVersion':
        return PromptChainVersion(
            id=data['id'],
            version=data['version'],
            promptChainId=data['promptChainId'],
            description=data.get('description'),
            config=PromptChainVersionConfig.from_dict(data['config']) if data.get('config') else None,
            createdAt=data['createdAt'],
            updatedAt=data['updatedAt']
        )
@dataclass
class PromptChainRuleType():
    field: str
    value: Union[str, int, List[str], bool, None]  # adding None here
    operator: str
    valueSource: Optional[str] = None
    exactMatch: Optional[bool] = None

    @staticmethod
    def from_dict(obj: Dict):
        return PromptChainRuleType(field=obj['field'], value=obj['value'], operator=obj['operator'], valueSource=obj.get('valueSource', None), exactMatch=obj.get('exactMatch', None))


@dataclass
class PromptChainRuleGroupType():
    rules: List[Union['PromptChainRuleType', 'PromptChainRuleGroupType']]
    combinator: str

    @staticmethod
    def from_dict(obj: Dict):
        rules = []
        for rule in obj['rules']:
            if 'rules' in rule:
                rules.append(PromptChainRuleGroupType.from_dict(rule))
            else:
                rules.append(PromptChainRuleType(**rule))
        return PromptChainRuleGroupType(rules=rules, combinator=obj['combinator'])


@dataclass
class PromptChainDeploymentRules():
    version: int
    query: Optional[PromptChainRuleGroupType] = None

    @staticmethod
    def from_dict(obj: Dict):
        query = obj.get('query', None)
        if query:
            query = PromptChainRuleGroupType.from_dict(query)
        return PromptChainDeploymentRules(version=obj['version'], query=query)
    
@dataclass
class VersionSpecificDeploymentConfig():
    id: str
    timestamp: datetime
    rules: PromptChainDeploymentRules
    isFallback: bool = False

    @staticmethod
    def from_dict(obj: Dict):
        rules = PromptChainDeploymentRules.from_dict(obj['rules'])
        return VersionSpecificDeploymentConfig(id=obj['id'], timestamp=obj['timestamp'], rules=rules, isFallback=obj.get('isFallback', False))
        
@dataclass
class PromptChainVersionsAndRules:
    folderId: str
    rules: Dict[str, List[VersionSpecificDeploymentConfig]]
    # rules: Dict[str, Union[str, int, bool, None]];
    versions: List[PromptChainVersion]
    fallbackVersion: Optional[PromptChainVersion]

    @staticmethod
    def from_dict(obj: Dict):
        rules = obj['rules']
        # Decoding each rule
        for key in rules:
            rules[key] = [VersionSpecificDeploymentConfig.from_dict(
                rule) for rule in rules[key]]
        versions = [PromptChainVersion.from_dict(version)
                    for version in obj['versions']]
        fallbackVersion = obj.get('fallbackVersion', None)
        if fallbackVersion:
            fallbackVersion = PromptChainVersion.from_dict(fallbackVersion)
        return PromptChainVersionsAndRules(rules=rules, versions=versions,  folderId=obj.get('folderId', None), fallbackVersion=fallbackVersion)

@ dataclass
class VersionAndRulesWithPromptChainId(PromptChainVersionsAndRules):
    promptChainId: str = ""

    @ staticmethod
    def from_dict(obj: Dict):
        rules = obj['rules']
        # Decoding each rule
        for key in rules:
            rules[key] = [VersionAndRulesWithPromptChainId(
                **rule) for rule in rules[key]]
        return VersionAndRulesWithPromptChainId(rules=rules, versions=obj['versions'], promptChainId=obj['promptChainId'], folderId=obj.get('folderId', None), fallbackVersion=obj.get('fallbackVersion', None))

@dataclass
class MaximApiPromptChainResponse:
    data: PromptChainVersionsAndRules
    error: Optional[dict]

    @staticmethod
    def from_dict(data: Dict) -> 'MaximApiPromptChainResponse':
        return MaximApiPromptChainResponse(
            data=PromptChainVersionsAndRules.from_dict(data['data']),
            error=data.get('error')
        )

@dataclass
class PromptChainWithId(PromptChainVersionsAndRules):
    promptChainId: str

class VersionAndRulesWithPromptChainIdEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, VersionAndRulesWithPromptChainId):
            return asdict(obj)
        return super().default(obj)
    
@dataclass
class MaximApiPromptChainsResponse:
    data: List[PromptChainWithId]
    error: Optional[dict]

    @staticmethod
    def from_dict(data: Dict) -> 'MaximApiPromptChainsResponse':
        return MaximApiPromptChainsResponse(
            data=[PromptChainWithId(**item) for item in data['data']],
            error=data.get('error')
        )


from typing import List, Optional


class MaximInMemoryCache():
    def __init__(self):
        self.cache = {}

    def get_all_keys(self) -> List[str]:
        return list(self.cache.keys())

    def get(self, key: str) -> Optional[str]:
        return self.cache.get(key)

    def set(self, key: str, value: str) -> None:
        self.cache[key] = value

    def delete(self, key: str) -> None:
        if key in self.cache:
            del self.cache[key]



import asyncio
import logging
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass, field
from functools import wraps
from typing import List, Optional
from uuid import uuid4

from ..logger.components.span import Span, SpanConfig
from ..logger.logger import Logger
from .trace import current_logger, current_trace


@dataclass
class _SpanStack:
    _stack: List[Span] = field(default_factory=list)

    def push(self, span: Span) -> None:
        self._stack.append(span)

    def pop(self) -> Optional[Span]:
        return self._stack.pop() if self._stack else None

    def current(self) -> Optional[Span]:
        return self._stack[-1] if self._stack else None


_span_ctx_var: ContextVar[_SpanStack] = ContextVar(
    "maxim_ctx_span_stack", default=_SpanStack([])
)


def current_span() -> Optional[Span]:
    return _span_ctx_var.get().current()


@contextmanager
def _push_span(span: Span):
    stack = _span_ctx_var.get()
    token = _span_ctx_var.set(_SpanStack([*stack._stack, span]))
    try:
        yield span
    finally:
        _span_ctx_var.reset(token)


def span(
    logger: Optional[Logger] = None,
    id: Optional[str] or Optional[Callable] = None,
    trace_id: Optional[str] or Optional[Callable] = None,
    name: Optional[str] = None,
    tags: Optional[dict] = None,
):
    """
    Decorator for creating a span within a trace.

    This decorator should be used within a function that is already decorated with @trace.
    It creates a new span and injects a tracer object into the decorated function.

    Args:
        logger (Logger): The Logger instance to use for logging.
        id (Optional[str] or Optional[Callable], optional): The ID for the span. If callable, it will be called to generate the ID. Defaults to None.
        trace_id (Optional[str] or Optional[Callable], optional): The trace ID to associate with this span. If callable, it will be called to generate the trace ID. Defaults to current_trace.
        name (Optional[str], optional): The name of the span. Defaults to None.
        tags (Optional[dict], optional): Additional tags to associate with the span. Defaults to None.

    Returns:
        Callable: A decorator function that wraps the original function with span functionality.

    Raises:
        ValueError: If the decorator is used outside of a @trace decorated function.
    """

    def decorator(func):

        if asyncio.iscoroutinefunction(func):

            @wraps(func)
            async def wrapper(*args, **kwargs):
                # First check if the logger is available
                maxim_logger = logger
                if maxim_logger is None:
                    if current_logger() is None:
                        raise ValueError(
                            "[MaximSDK]: no logger found. either call this function from a @trace decorated function or pass a logger"
                        )
                    maxim_logger = current_logger()
                # Checking if its a sub-span
                parent_span = current_span()
                # We will try to get the current trace id
                actual_trace_id = trace_id() if callable(trace_id) else trace_id
                if actual_trace_id is None:
                    actual_trace_id = current_trace().id
                if actual_trace_id is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace_id found. either call this function from a @trace decorated function or pass a trace_id"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace_id found. either call this function from a @trace decorated function or pass a trace_id"
                        )
                actual_span_id = id() if callable(id) else id
                if actual_span_id is None:
                    actual_span_id = str(uuid4())
                # Creating span
                span_config = SpanConfig(id=actual_span_id, name=name, tags=tags)
                span: Span
                if parent_span is not None:
                    span = parent_span.span(span_config)
                else:
                    span = maxim_logger.trace_add_span(
                        actual_trace_id, config=span_config
                    )
                # If actual_trace_id is None, we will try to get the current trace id
                try:
                    with _push_span(span):
                        return await func(*args, **kwargs)
                finally:
                    span.end()

        else:

            @wraps(func)
            def wrapper(*args, **kwargs):
                # First check if the logger is available
                maxim_logger = logger
                if maxim_logger is None:
                    if current_logger() is None:
                        raise ValueError(
                            "[MaximSDK]: no logger found. either call this function from a @trace decorated function or pass a logger"
                        )
                    maxim_logger = current_logger()
                # Checking if its a sub-span
                parent_span = current_span()
                # We will try to get the current trace id
                actual_trace_id = trace_id() if callable(trace_id) else trace_id
                if actual_trace_id is None:
                    actual_trace_id = current_trace().id
                if actual_trace_id is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace_id found. either call this function from a @trace decorated function or pass a trace_id"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace_id found. either call this function from a @trace decorated function or pass a trace_id"
                        )
                actual_span_id = id() if callable(id) else id
                if actual_span_id is None:
                    actual_span_id = str(uuid4())
                # Creating span
                span_config = SpanConfig(id=actual_span_id, name=name, tags=tags)
                span: Span
                if parent_span is not None:
                    span = parent_span.span(span_config)
                else:
                    span = maxim_logger.trace_add_span(
                        actual_trace_id, config=span_config
                    )
                # If actual_trace_id is None, we will try to get the current trace id
                try:
                    with _push_span(span):
                        return func(*args, **kwargs)
                finally:
                    span.end()

        return wrapper

    return decorator



from typing import Dict, Optional, Union


class VariableType(str):
    TEXT = "text"
    JSON = "json"
    

class Variable():
    def __init__(self, type: str, payload: Dict[str, Union[str, int, bool]]):
        self.type = type
        self.payload = payload

    def to_json(self):
        return {
            "type": self.type,
            "payload": self.payload
        }

class DatasetEntry():
    def __init__(self, input: Variable, context: Optional[Variable] = None, expectedOutput: Optional[Variable] = None):
        self.input = input
        self.context = context
        self.expectedOutput = expectedOutput

    def to_json(self):
        return_dict = {}
        if self.input is not None:
            return_dict["input"] = {
                "type": self.input['type'],  # type: ignore
                "payload": self.input['payload']  # type: ignore
            }
        if self.context is not None:
            return_dict["context"] = {
                "type": self.context['type'], # type: ignore
                "payload": self.context['payload'] # type: ignore
            }
        if self.expectedOutput is not None:
            return_dict["expectedOutput"] = {
                "type": self.expectedOutput['type'], # type: ignore
                "payload": self.expectedOutput['payload'] # type: ignore
            }
        return return_dict
        


from dataclasses import dataclass
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Dict, Optional

from langchain_core.messages import tool

from ..writer import LogWriter
from .base import EventEmittingBaseContainer
from .feedback import Feedback
from .generation import Generation, GenerationConfig
from .retrieval import Retrieval, RetrievalConfig
from .toolCall import ToolCall, ToolCallConfig
from .types import Entity

if TYPE_CHECKING:
    from .span import Span, SpanConfig  # Type checking only


@dataclass
class TraceConfig:
    id: str
    name: Optional[str] = None
    session_id: Optional[str] = None
    tags: Optional[Dict[str, str]] = None
    input: Optional[str] = None


class Trace(EventEmittingBaseContainer):

    def __init__(self, config: TraceConfig, writer: LogWriter):
        self.output = None
        super().__init__(Entity.TRACE, config.__dict__, writer)
        payload_to_send = {
            **self.data(),
            "sessionId": config.session_id,
        }
        if config.input is not None:
            payload_to_send["input"] = config.input
        self._commit("create", payload_to_send)

    def set_input(self, input: str):
        self._commit("update", {"input": input})

    @staticmethod
    def set_input_(writer: LogWriter, trace_id: str, input: str):
        Trace._commit_(writer, Entity.TRACE, trace_id, "update", {"input": input})

    def set_output(self, output: str):
        self.output = output
        self._commit("update", {"output": output})

    @staticmethod
    def set_output_(writer: LogWriter, trace_id: str, output: str):
        Trace._commit_(writer, Entity.TRACE, trace_id, "update", {"output": output})

    def generation(self, config: GenerationConfig) -> Generation:
        generation = Generation(config, self.writer)
        self._commit(
            "add-generation",
            {
                **generation.data(),
                "id": generation.id,
            },
        )
        return generation

    def tool_call(self, config: ToolCallConfig) -> ToolCall:
        tool_call = ToolCall(config, self.writer)
        self._commit(
            "add-tool-call",
            {
                **tool_call.data(),
                "id": tool_call.id,
            },
        )
        return tool_call

    @staticmethod
    def tool_call_(
        writer: LogWriter, trace_id: str, config: ToolCallConfig
    ) -> ToolCall:
        tool_call = ToolCall(config, writer)
        Trace._commit_(
            writer,
            Entity.TRACE,
            trace_id,
            "add-tool-call",
            {
                **tool_call.data(),
                "id": tool_call.id,
            },
        )
        return tool_call

    @staticmethod
    def generation_(
        writer: LogWriter, trace_id: str, config: GenerationConfig
    ) -> Generation:
        generation = Generation(config, writer)
        Trace._commit_(
            writer,
            Entity.TRACE,
            trace_id,
            "add-generation",
            {
                **generation.data(),
                "id": generation.id,
            },
        )
        return generation

    def retrieval(self, config: RetrievalConfig):
        retrieval = Retrieval(config, self.writer)
        self._commit(
            "add-retrieval",
            {
                "id": config.id,
                **retrieval.data(),
            },
        )
        return retrieval

    @staticmethod
    def retrieval_(writer: LogWriter, trace_id: str, config: RetrievalConfig):
        retrieval = Retrieval(config, writer)
        Trace._commit_(
            writer,
            Entity.TRACE,
            trace_id,
            "add-retrieval",
            {
                "id": config.id,
                **retrieval.data(),
            },
        )
        return retrieval

    def span(self, config: "SpanConfig") -> "Span":
        from .span import Span, SpanConfig

        span = Span(config, self.writer)
        self._commit(
            "add-span",
            {
                "id": config.id,
                **span.data(),
            },
        )
        return span

    @staticmethod
    def span_(writer: LogWriter, trace_id: str, config: "SpanConfig") -> "Span":
        from .span import Span, SpanConfig

        span = Span(config, writer)
        Trace._commit_(
            writer,
            Entity.TRACE,
            trace_id,
            "add-span",
            {
                "id": config.id,
                **span.data(),
            },
        )

        return span

    def feedback(self, feedback: Feedback):
        self._commit("add-feedback", feedback.__dict__)

    @staticmethod
    def feedback_(writer: LogWriter, trace_id: str, feedback: Feedback):
        Trace._commit_(
            writer, Entity.TRACE, trace_id, "add-feedback", feedback.__dict__
        )

    @staticmethod
    def add_tag_(writer: LogWriter, id: str, key: str, value: str):
        EventEmittingBaseContainer._add_tag_(writer, Entity.TRACE, id, key, value)

    @staticmethod
    def end_(writer: LogWriter, trace_id: str, data: Optional[Dict[str, str]] = None):
        if data is None:
            data = {}
        return EventEmittingBaseContainer._end_(
            writer,
            Entity.TRACE,
            trace_id,
            {
                "endTimestamp": datetime.now(timezone.utc),
                **data,
            },
        )

    @staticmethod
    def event_(
        writer: LogWriter,
        trace_id: str,
        id: str,
        event: str,
        tags: Optional[Dict[str, str]] = None,
    ):
        return EventEmittingBaseContainer._event_(
            writer, Entity.TRACE, trace_id, id, event, tags
        )

    def data(self) -> Dict[str, Any]:
        return {
            **super().data(),
            "output": self.output,
        }



import logging
import os
import tempfile
import threading
import time
import uuid
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from typing import Optional, Union

from ..apis.maxim import MaximAPI

maximLogger = logging.getLogger("MaximSDK")


class LogWriterConfig:
    def __init__(self, base_url, api_key, repository_id, auto_flush=True, flush_interval: Optional[int] = 10, is_debug=False, raise_exceptions=False):
        self.base_url = base_url
        self.api_key = api_key
        self.repository_id = repository_id
        self.auto_flush = auto_flush
        self.flush_interval = flush_interval
        self.is_debug = is_debug
        self.raise_exceptions = raise_exceptions
        maximLogger.setLevel(logging.DEBUG if is_debug else logging.INFO)

class LogWriter:
    def __init__(self, config: LogWriterConfig):
        self.is_running = True
        self.id = str(uuid.uuid4())
        self.config = config
        self.maxim_api = MaximAPI(config.base_url, config.api_key)
        self.queue = Queue()
        self.executor = ThreadPoolExecutor(max_workers=1)
        self.is_debug = config.is_debug
        self.raise_exceptions = config.raise_exceptions
        self.logs_dir = os.path.join(
            tempfile.gettempdir(), f"maxim-sdk/{self.id}/maxim-logs")
        self.__flush_thread = None
        try:
            os.makedirs(self.logs_dir, exist_ok=True)
        except Exception as ignored:
            maximLogger.info("Maxim library does not have FS access")
        if self.config.auto_flush:
            if self.config.flush_interval:
                maximLogger.debug(
                    f"Starting flush thread with interval {self.config.flush_interval} seconds")
                self.__flush_thread = threading.Thread(target=self.__sync_timer)
                self.__flush_thread.daemon = True
                self.__flush_thread.start()
            else:
                raise ValueError(
                    "flush_interval is set to None.flush_interval has to be a number")

    def __sync_timer(self):
        while self.is_running:
            self.flush()
            if self.config.flush_interval is None:
                raise ValueError("flush_interval is set to None. flush_interval has to be a number")
            time.sleep(self.config.flush_interval)

    def is_running_on_lambda(self):
        return 'AWS_LAMBDA_FUNCTION_NAME' in os.environ

    def write_to_file(self, logs):
        try:
            filename = f"logs-{time.strftime('%Y-%m-%dT%H:%M:%SZ')}.log"
            filepath = os.path.join(self.logs_dir, filename)
            maximLogger.debug(f"Writing logs to file: {filename}")
            with open(filepath, 'w') as file:
                for log in logs:
                    file.write(log.serialize() + "\n")
            return filepath
        except Exception as e:
            maximLogger.error(f"Failed to write logs to file. We will keep it in memory. Error: {e}")
            if self.raise_exceptions:
                raise e
            return None

    def flush_log_files(self):
        try:
            if os.path.exists(self.logs_dir) == False:
                return
            files = os.listdir(self.logs_dir)
            for file in files:
                with open(os.path.join(self.logs_dir, file), 'r') as f:
                    logs = f.read()
                try:
                    self.maxim_api.pushLogs( self.config.repository_id, logs)
                    os.remove(os.path.join(self.logs_dir, file))
                except Exception as e:
                    maximLogger.error(
                        f"Failed to access filesystem. Error: {e}")
                    if self.raise_exceptions:
                        raise Exception(e)
        except Exception as e:
            maximLogger.error(f"Failed to access filesystem. Error: {e}")
            if self.raise_exceptions:
                raise Exception(e)

    def can_access_filesystem(self):
        try:
            return os.access(tempfile.gettempdir(), os.W_OK)
        except Exception as e:
            return False

    def flush_logs(self, logs):
        try:
            # Pushing old logs first
            if self.can_access_filesystem():
                self.flush_log_files()
            # Pushing new logs
            logs_to_push = "\n".join(
                [log.serialize() for log in logs])
            self.maxim_api.pushLogs(self.config.repository_id, logs_to_push)
            maximLogger.debug("Flush complete")
        except Exception as e:
            if self.is_running_on_lambda():
                maximLogger.debug("As we are running on lamda - we will keep logs in memory for next attempt")
                for log in logs:
                    self.queue.put(log)
                    maximLogger.debug("Logs added back to queue for next attempt")
            else:
                if self.can_access_filesystem():
                    self.write_to_file(logs)
                    maximLogger.warning(f"Failed to push logs to server. Writing logs to file. Error: {e}")
                else:
                    for log in logs:
                        self.queue.put(log)
                        maximLogger.debug("Logs added back to queue for next attempt")

    def commit(self, log):
        self.queue.put(log)

    def flush(self):
        items = []
        while not self.queue.empty():
            items.append(self.queue.get())
        if len(items) == 0:
            self.flush_log_files()
            maximLogger.debug("No logs to flush")
            return
        maximLogger.debug(
            f"Flushing logs to server {time.strftime('%Y-%m-%dT%H:%M:%S')}")
        if self.is_debug:
            for item in items:
                maximLogger.debug(f"{item.serialize()}")
        self.executor.submit(self.flush_logs, items)

    def cleanup(self):
        maximLogger.debug("Cleaning up writer")
        self.is_running = False
        self.flush()
        maximLogger.debug("Waiting for executor to shutdown")
        self.executor.shutdown(wait=True)
        maximLogger.debug("Writer cleanup complete")



from .components.feedback import Feedback
from .components.generation import Generation, GenerationConfig, GenerationError
from .components.retrieval import Retrieval, RetrievalConfig
from .components.session import Session, SessionConfig
from .components.span import Span, SpanConfig
from .components.toolCall import ToolCall, ToolCallConfig
from .components.trace import Trace, TraceConfig
from .logger import Logger, LoggerConfig

__all__ = [
    "Logger",
    "LoggerConfig",
    "TraceConfig",
    "Trace",
    "RetrievalConfig",
    "Retrieval",
    "SpanConfig",
    "Span",
    "GenerationConfig",
    "Generation",
    "SessionConfig",
    "Session",
    "Feedback",
    "ToolCallConfig",
    "ToolCall",
]



import asyncio
import json
from contextvars import ContextVar
from functools import wraps
from typing import Callable, Dict, Optional
from uuid import uuid4

from ..logger.components.retrieval import Retrieval, RetrievalConfig
from ..logger.logger import Logger
from .span import current_span
from .trace import current_logger, current_trace

_retrieval_ctx_var: ContextVar[Optional[Retrieval]] = ContextVar(
    "maxim_ctx_retrieval", default=None
)


def current_retrieval() -> Optional[Retrieval]:
    return _retrieval_ctx_var.get()


def retrieval(
    logger: Optional[Logger] = None,
    id: Optional[str] or Optional[Callable] = None,
    input: Optional[str] or Optional[Callable] = None,
    name: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
):
    def decorator(func):        
        if asyncio.iscoroutinefunction(func):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                # First check if the logger is available
                maxim_logger = logger
                if maxim_logger is None:
                    if current_logger() is None:
                        raise ValueError(
                            "[MaximSDK]: no logger found. either call this function from a @trace decorated function or pass a logger"
                        )
                    maxim_logger = current_logger()
                # Here there should be an active span or active trace
                # If none of this is the case then we raise an error
                if current_span() is None and current_trace() is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                actual_retrieval_id = id() if callable(id) else id
                if actual_retrieval_id is None:
                    actual_retrieval_id = str(uuid4())
                config = RetrievalConfig(id=actual_retrieval_id, name=name, tags=tags)
                retrieval: Retrieval = None
                if current_span() is not None:
                    retrieval = current_span().retrieval(config)
                elif current_trace() is not None:
                    retrieval = current_trace().retrieval(config)
                # Checking for input
                actual_input = input() if callable(input) else input
                if actual_input is not None:
                    retrieval.input(actual_input)
                token = _retrieval_ctx_var.set(retrieval)
                try:
                    result = await func(*args, **kwargs)
                    # Here we will check if retrieval has output
                    if not retrieval.is_output_set:
                        if isinstance(result, (dict)):
                            retrieval.output(json.dumps(result))
                        elif isinstance(result, list):
                            retrieval.output([json.dumps(item) for item in result])
                        elif isinstance(result, str):
                            retrieval.output(result)
                    return result
                finally:
                    retrieval.end()
                    _retrieval_ctx_var.reset(token)
        else:
            @wraps(func)
            def wrapper(*args, **kwargs):
                # First check if the logger is available
                maxim_logger = logger
                if maxim_logger is None:
                    if current_logger() is None:
                        raise ValueError(
                            "[MaximSDK]: no logger found. either call this function from a @trace decorated function or pass a logger"
                        )
                    maxim_logger = current_logger()
                # Here there should be an active span or active trace
                # If none of this is the case then we raise an error
                if current_span() is None and current_trace() is None:
                    if maxim_logger.raise_exceptions:
                        raise ValueError(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                    else:
                        logging.warning(
                            "[MaximSDK]: no trace or span found. either call this function from a @trace or @span decorated function"
                        )
                actual_retrieval_id = id() if callable(id) else id
                if actual_retrieval_id is None:
                    actual_retrieval_id = str(uuid4())
                config = RetrievalConfig(id=actual_retrieval_id, name=name, tags=tags)
                retrieval: Retrieval = None
                if current_span() is not None:
                    retrieval = current_span().retrieval(config)
                elif current_trace() is not None:
                    retrieval = current_trace().retrieval(config)
                # Checking for input
                actual_input = input() if callable(input) else input
                if actual_input is not None:
                    retrieval.input(actual_input)
                token = _retrieval_ctx_var.set(retrieval)
                try:
                    result = func(*args, **kwargs)
                    # Here we will check if retrieval has output
                    if not retrieval.is_output_set:
                        if isinstance(result, (dict)):
                            retrieval.output(json.dumps(result))
                        elif isinstance(result, list):
                            retrieval.output([json.dumps(item) for item in result])
                        elif isinstance(result, str):
                            retrieval.output(result)
                    return result
                finally:
                    retrieval.end()
                    _retrieval_ctx_var.reset(token)
        return wrapper

    return decorator



from .generation import current_generation, generation
from .retrieval import current_retrieval, retrieval
from .span import current_span, span
from .trace import current_trace, trace

__all__ = [
    "trace",
    "span",
    "current_trace",
    "current_span",
    "current_retrieval",
    "retrieval",
    "current_generation",
    "generation",
]



from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, Optional

from ..writer import LogWriter
from .base import EventEmittingBaseContainer
from .feedback import Feedback
from .trace import Trace, TraceConfig
from .types import Entity


@dataclass
class SessionConfig:
    id: str
    name: Optional[str] = None
    tags: Optional[Dict[str, str]] = None


class Session(EventEmittingBaseContainer):
    ENTITY = Entity.SESSION

    def __init__(self, config: SessionConfig, writer: LogWriter):
        super().__init__(Session.ENTITY, config.__dict__, writer)
        self._commit("create")

    def trace(self, config: TraceConfig) -> Trace:
        # Assuming TraceConfig in Python has a session_id attribute
        config.session_id = self.id
        return Trace(config, self.writer)

    @staticmethod
    def trace_(writer: LogWriter, session_id: str, config: TraceConfig) -> Trace:
        config.session_id = session_id
        return Trace(config, writer)

    def feedback(self, feedback: Feedback):
        self._commit("add-feedback", feedback.__dict__)

    @staticmethod
    def feedback_(writer: LogWriter, session_id: str, feedback: Feedback):
        EventEmittingBaseContainer._commit_(writer,
                                            Entity.SESSION, session_id, "add-feedback", feedback.__dict__)

    @staticmethod
    def add_tag_(writer: LogWriter, session_id: str, key: str, value: str):
        return EventEmittingBaseContainer._add_tag_(writer, Entity.SESSION, session_id, key, value)

    @staticmethod
    def end_(writer: LogWriter, session_id: str, data: Optional[Dict[str, Any]] = None):
        if data is None:
            data = {}
        return EventEmittingBaseContainer._end_(writer, Entity.SESSION, session_id, {
            "endTimestamp": datetime.now(timezone.utc),
            **data,
        })

    @staticmethod
    def event_(writer: LogWriter, session_id: str, id: str, event: str, data: Dict[str, str]):
        return EventEmittingBaseContainer._event_(writer, Entity.SESSION, session_id, id, event, data)



import codecs
import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = "3.0.0rc6"
DESCRIPTION = 'Maxim Python Library'
LONG_DESCRIPTION = 'A package that allows you to use the Maxim Python Library to interact with the Maxim Platform'

print("=============================")
print("Releasing version: ", VERSION)
print("=============================")

packages = find_packages()
packages.remove('maxim.tests')
# Setting up
setup(
    name="maxim-py",
    version=VERSION,
    author="Maxim Engineering",
    author_email="<eng@getmaxim.ai>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=packages,
    install_requires=[],
    extras_require={},
    keywords=['python', 'prompts', 'logs', 'workflow', 'testing'],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



import contextlib
import json
import logging
from typing import Any, Dict, List, Optional

import requests
from requests.adapters import HTTPAdapter
from urllib3.poolmanager import PoolManager
from urllib3.util import Retry, Timeout

from ..models.dataset import DatasetEntry
from ..models.folder import Folder
from ..models.prompt import (Prompt, VersionAndRulesWithPromptId,
                             VersionsAndRules)
from ..models.promptChain import VersionAndRulesWithPromptChainId


class ConnectionPool:
    def __init__(self):
        self.session = requests.Session()
        retries = Retry(connect=5, read=3, redirect=1, status=3, backoff_factor=0.4,
                        status_forcelist=frozenset({413, 429, 500, 502, 503, 504}))
        self.http = PoolManager(num_pools=2, maxsize=3, retries=retries, timeout=Timeout(connect=10, read=10))
        self.session.mount('https://', HTTPAdapter(max_retries=retries))
        self.session.mount('http://', HTTPAdapter(max_retries=retries))

    @contextlib.contextmanager
    def get_session(self):
        try:
            yield self.session
        finally:
            self.session.close()

    @contextlib.contextmanager
    def get_connection(self):
        try:
            yield self.http
        finally:
            self.http.clear()


class MaximAPI:
    connection_pool: ConnectionPool

    def __init__(self,base_url: str, api_key: str):
        self.connection_pool = ConnectionPool()
        self.base_url = base_url
        self.api_key = api_key

    def __make_network_call(self, method: str, endpoint: str, body: Optional[str] = None) -> bytes:
        headers = {"x-maxim-api-key": self.api_key}
        url = f"{self.base_url}{endpoint}"
        with self.connection_pool.get_session() as session:
            response = session.request(
                method,
                url,
                data=body,
                headers=headers
            )
            response.raise_for_status()
            return response.content

    def get_prompt(self, id: str) -> VersionAndRulesWithPromptId:
        res = self.__make_network_call(
            "GET", f"/api/sdk/v4/prompts?promptId={id}")
        data = json.loads(res.decode())['data']
        return VersionAndRulesWithPromptId.from_dict(data)

    def get_prompts(self) -> List[VersionAndRulesWithPromptId]:
        res = self.__make_network_call("GET", "/api/sdk/v4/prompts")
        return [VersionAndRulesWithPromptId.from_dict(data) for data in json.loads(res)['data']]

    def getPromptChain(self, id: str) -> VersionAndRulesWithPromptChainId:
        res = self.__make_network_call(
            "GET", f"/api/sdk/v4/prompt-chains?promptChainId={id}")
        json_response = json.loads(res.decode())
        return VersionAndRulesWithPromptChainId(**json_response['data'])

    def get_prompt_chains(self) -> List[VersionAndRulesWithPromptChainId]:
        res = self.__make_network_call("GET", "/api/sdk/v4/prompt-chains")
        json_response = json.loads(res.decode())
        return [VersionAndRulesWithPromptChainId(**elem) for elem in json_response['data']]

    def get_folder(self, id: str) -> Folder:
        res = self.__make_network_call("GET", f"/api/sdk/v3/folders?folderId={id}")
        json_response = json.loads(res.decode())
        if 'tags' not in json_response:
            json_response['tags'] = {}
        return Folder(**json_response['data'])

    def get_folders(self) -> List[Folder]:
        res = self.__make_network_call("GET", "/api/sdk/v3/folders")
        json_response = json.loads(res.decode())
        for elem in json_response['data']:
            if 'tags' not in elem:
                elem['tags'] = {}
        return [Folder(**elem) for elem in json_response['data']]

    def addDatasetEntries(self,dataset_id: str, dataset_entries: List[DatasetEntry]) -> dict:
        res = self.__make_network_call("POST", "/api/sdk/v3/datasets/entries", json.dumps({"datasetId": dataset_id, "entries": [entry.to_json() for entry in dataset_entries]}))
        return json.loads(res.decode())

    def does_log_repository_exist(self, logger_id: str) -> bool:
        try:
            res = self.__make_network_call("GET", f"/api/sdk/v3/log-repositories?loggerId={logger_id}")
            json_response = json.loads(res.decode())
            if 'error' in json_response:
                return False
            return True
        except Exception as e:
            return False

    def pushLogs(self, repository_id: str, logs: str) -> None:
        try:
            res = self.__make_network_call("POST", f"/api/sdk/v3/log?id={repository_id}", logs)
            json_response = json.loads(res.decode())
            if 'error' in json_response:
                raise Exception(json_response['error'])
        except Exception as e:
            raise Exception(e)



import logging
import sys
from bdb import GENERATOR_AND_COROUTINE_FLAGS
from configparser import LegacyInterpolation
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Sequence, Union
from uuid import UUID, uuid4

import maxim

# Use LangChain features
from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.documents import Document
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    ChatMessage,
    FunctionMessage,
    HumanMessage,
    SystemMessage,
    ToolMessage,
)
from langchain_core.outputs import LLMResult
from maxim import expiringKeyValueStore
from maxim.logger.langchain.utils import parse_base_message_to_maxim_generation

from ...expiringKeyValueStore import ExpiringKeyValueStore
from ...logger import (
    GenerationConfig,
    Logger,
    RetrievalConfig,
    Span,
    SpanConfig,
    ToolCall,
    ToolCallConfig,
    TraceConfig,
)
from .utils import (parse_langchain_llm_error, parse_langchain_llm_result,
                    parse_langchain_messages,
                    parse_langchain_model_and_provider,
                    parse_langchain_model_parameters, parse_langchain_provider)

logger = logging.getLogger("MaximSDK")


# 20 minutes
DEFAULT_TIMEOUT = 60 * 20


class ChainType(Enum):
    CHAIN = "chain"
    GENERATION = "generation"
    TOOL = "tool"


class MaximLangchainTracer(BaseCallbackHandler):
    """A callback handler that logs langchain outputs to Maxim logger

    Args:
        logger: Logger: Maxim Logger instance to log outputs
    """

    def __init__(self, logger: Logger, metadata: Optional[Dict[str, Any]]) -> None:
        """Initializes the Langchain Tracer
        Args:
            logger: Logger: Maxim Logger instance to log outputs
        """
        super().__init__()
        self.logger = logger
        self.chains = ExpiringKeyValueStore()
        self.containers = ExpiringKeyValueStore()
        self.tool_calls = ExpiringKeyValueStore()
        self.generations = ExpiringKeyValueStore()
        self.retrievals = ExpiringKeyValueStore()
        self.metadata_store = ExpiringKeyValueStore()
        self.metadata = None
        if metadata is not None:
            self._validate_maxim_metadata(metadata)
            self.metadata = metadata

    def _get_chain_type(self, messages: List[Any]) -> ChainType:
        # Checking if there is already a generation running for this run_id
        if len(messages) == 0:
            return ChainType.CHAIN
        # Now we will check if the messages are type of one of the messages
        try:
            if any(isinstance(message, FunctionMessage) for message in messages) or any(
                isinstance(message, ToolMessage) for message in messages
            ):
                return ChainType.TOOL
            if any(
                isinstance(
                    message, (HumanMessage, AIMessage, ChatMessage, SystemMessage)
                )
                for message in messages
            ):
                return ChainType.GENERATION

            return ChainType.CHAIN
        except Exception as e:
            return ChainType.CHAIN

    def _validate_maxim_metadata(self, metadata: Dict[str, Any]):
        if metadata is None:
            return
        id_keys = ["session_id", "trace_id", "span_id"]
        present_keys = [key for key in id_keys if key in metadata]
        if len(present_keys) > 1:
            raise ValueError(
                f"Multiple keys found in metadata: {present_keys}. You can pass only one of these."
            )
        valid_keys = [
            "session_id",
            "trace_id",
            "span_id",
            "generation_tags",
            "retrieval_tags",
            "trace_tags",
            "retrieval_name",
            "trace_name",
            "generation_name",
        ]
        invalid_keys = [key for key in metadata if key not in valid_keys]
        if len(invalid_keys) > 0:
            raise ValueError(
                f"Invalid keys found in metadata: {invalid_keys}. Valid keys are {valid_keys}"
            )

    def _get_container(self, run_id: UUID):
        # Now we will checking in containers
        if self.containers.get(str(run_id)) is not None:
            container = self.containers.get(str(run_id))
            if container.startswith("span:"):
                return container.split(":", 1)
            if container.startswith("trace:"):
                return container.split(":", 1)
            if container.startswith("session:"):
                return container.split(":", 1)
            return None, None
        return None, None

    def _get_container_from_metadata(self, run_id: UUID):
        maxim_metadata = self.metadata_store.get(str(run_id))
        if maxim_metadata is not None:
            span_id = maxim_metadata.get("span_id", None)
            if span_id is not None:
                return "span", span_id
            session_id = maxim_metadata.get("session_id", None)
            if session_id is not None:
                return "session", session_id
            trace_id = maxim_metadata.get("trace_id", None)
            if trace_id is not None:
                return "trace", trace_id
        return "local_trace", str(run_id)

    # LLM callbacks

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: UUID,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"[MaximSDK: Langchain] on_llm_start called")
        model, model_parameters = parse_langchain_model_parameters(**kwargs)
        provider = parse_langchain_provider(serialized)
        model, provider = parse_langchain_model_and_provider(model, provider)
        messages = parse_langchain_messages(prompts)
        generation_id = str(uuid4())
        generation_name = None
        generation_tags = None
        trace_tags = None

        if metadata is not None and metadata.get("maxim", None) is not None:
            self._validate_maxim_metadata(metadata.get("maxim", None))
            self.metadata_store.set(
                str(run_id), metadata.get("maxim", None), DEFAULT_TIMEOUT
            )
            trace_tags = metadata.get("maxim", {}).get("trace_tags", None)
            generation_name = metadata.get("maxim", {}).get("generation_name", None)
            generation_tags = metadata.get("maxim", {}).get("generation_tags", None)
        elif self.metadata is not None:
            self.metadata_store.set(str(run_id), self.metadata, DEFAULT_TIMEOUT)
            trace_name = self.metadata.get("trace_name", None)
            trace_tags = self.metadata.get("trace_tags", None)
            generation_name = self.metadata.get("generation_name", None)
            generation_tags = self.metadata.get("generation_tags", None)

        generation_config = GenerationConfig(
            id=generation_id,
            name=generation_name,
            provider=provider,
            model=model,
            messages=messages,
            model_parameters=model_parameters,
            tags=generation_tags,
        )

        self.generations.set(str(run_id), generation_id, DEFAULT_TIMEOUT)

        container, container_id = self._get_container_from_metadata(run_id)

        if container == "span":
            self.logger.span_generation(container_id, generation_config)
        elif container == "trace":
            self.logger.trace_generation(container_id, generation_config)
        elif container == "session":
            trace_name = None
            if metadata is not None:
                trace_name = metadata.get("maxim", {}).get("trace_name", None)
            trace = self.logger.session_trace(
                container_id,
                TraceConfig(id=str(run_id), name=trace_name, tags=trace_tags),
            )
            trace.generation(generation_config)
        else:
            trace_name = None
            if metadata is not None:
                trace_name = metadata.get("maxim", {}).get("trace_name", None)
            trace = self.logger.trace(
                TraceConfig(id=str(run_id), name=trace_name, tags=trace_tags)
            )
            trace.generation(generation_config)

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> Any:
        """Run when LLM ends running."""
        logger.debug(f"[MaximSDK: Langchain] on_llm_end called")
        run_id = kwargs.get("run_id", None)
        if run_id is not None:
            result = parse_langchain_llm_result(response)
            generation_id = self.generations.get(str(run_id))
            if generation_id is not None:
                container, container_id = self._get_container_from_metadata(run_id)
                self.logger.generation_result(generation_id, result)
                if container == "local_trace":
                    self.logger.trace_end(container_id)

    # Retrieval callbacks

    def on_retriever_start(
        self,
        serialized: Dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"[Langchain] on_retriever_start called")
        """Run when Retriever starts running."""
        container, container_id = self._get_container_from_metadata(run_id)
        retrieval_id = str(uuid4())
        retrieval_config = RetrievalConfig(id=retrieval_id)
        if container == "span":
            retrieval = self.logger.span_retrieval(
                container_id, config=retrieval_config
            )
            retrieval.input(query)
        elif container == "trace":
            retrieval = self.logger.trace_retrieval(
                container_id, config=retrieval_config
            )
            retrieval.input(query)
        elif container == "session":
            trace = self.logger.session_trace(container_id, TraceConfig(id=str(run_id)))
            retrieval = trace.retrieval(retrieval_config)
            retrieval.input(query)
        else:
            retrieval = self.logger.trace_retrieval(
                str(run_id), config=retrieval_config
            )
            retrieval.input(query)

    def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"[Langchain] on_retriever_end called")
        trace_id = str(run_id)
        retrieval_id = self.retrievals.get(trace_id)
        if retrieval_id is not None:
            self.logger.retrieval_output(retrieval_id, documents)

    # ChatModel callbacks

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"[Langchain] on_chat_model_start called")
        run_id = kwargs.get("run_id", None)
        model, model_parameters = parse_langchain_model_parameters(**kwargs)
        provider = parse_langchain_provider(serialized)
        model, provider = parse_langchain_model_and_provider(model, provider)
        maxim_messages = parse_langchain_messages(messages)
        generation_id = str(uuid4())
        # Checking if generation id already present for this run_id
        if self.generations.get(str(run_id)) is not None:
            generation_id = self.generations.get(str(run_id))
        generation_name = None
        generation_tags = None
        trace_tags = None
        trace_name = None

        if metadata is not None and metadata.get("maxim", None) is not None:
            self._validate_maxim_metadata(metadata.get("maxim", None))
            self.metadata_store.set(
                str(run_id), metadata.get("maxim", None), DEFAULT_TIMEOUT
            )
            trace_name = metadata.get("maxim", {}).get("trace_name", None)
            trace_tags = metadata.get("maxim", {}).get("trace_tags", None)
            generation_name = metadata.get("maxim", {}).get("generation_name", None)
            generation_tags = metadata.get("maxim", {}).get("generation_tags", None)
        elif self.metadata is not None:
            self.metadata_store.set(str(run_id), self.metadata, DEFAULT_TIMEOUT)
            trace_name = self.metadata.get("trace_name", None)
            trace_tags = self.metadata.get("trace_tags", None)
            generation_name = self.metadata.get("generation_name", None)
            generation_tags = self.metadata.get("generation_tags", None)

        generation_config = GenerationConfig(
            id=generation_id,
            name=generation_name,
            provider=provider,
            model=model,
            messages=maxim_messages,
            model_parameters=model_parameters,
            tags=generation_tags,
        )
        self.generations.set(str(run_id), generation_id, DEFAULT_TIMEOUT)

        container, container_id = self._get_container_from_metadata(run_id)

        if container == "span":
            self.logger.span_generation(container_id, generation_config)
        elif container == "trace":
            self.logger.trace_generation(container_id, generation_config)
        elif container == "session":
            trace = self.logger.session_trace(
                container_id,
                TraceConfig(id=str(run_id), name=trace_name, tags=trace_tags),
            )
            trace.generation(generation_config)
        else:
            trace = self.logger.trace(
                TraceConfig(id=str(run_id), name=trace_name, tags=trace_tags)
            )
            trace.generation(generation_config)

    def on_llm_new_token(self, token: str, **kwargs: Any) -> Any:
        """Run on new LLM token. Only available when streaming is enabled."""

    def on_llm_error(
        self, error: Union[Exception, BaseException, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        """Run when LLM errors."""
        logger.debug(f"[Langchain] on_llm_error called")
        run_id = kwargs.get("run_id", None)
        if run_id is not None:
            generation_id = self.generations.get(str(run_id))
            if generation_id is not None:
                container, container_id = self._get_container_from_metadata(run_id)
                generation_error = parse_langchain_llm_error(error)
                self.logger.generation_error(generation_id, generation_error)
                if container == "local_trace":
                    self.logger.trace_end(container_id)

    # Chain callbacks

    def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> Any:
        try:
            # We hide the hidden chains
            if "langsmith:hidden" in kwargs.get("tags", []):
                return
            parent_run_id = kwargs.get("parent_run_id", None)
            run_id = kwargs.get("run_id", None)
            span_name = kwargs.get("name", None)
            span: Optional[Span] = None
            span_id = str(uuid4())
            tags: Dict[str, str] = {}
            tags["run_id"] = str(run_id)
            if parent_run_id is not None:
                tags["parent_run_id"] = str(parent_run_id)
            if kwargs.get("tags", None) is not None:
                for tag in kwargs["tags"]:
                    key, value = tag.split(":", 1)
                    tags[key.strip()] = value.strip()
            metadata = kwargs.get("metadata", None)
            if metadata is not None:
                for key, value in kwargs["metadata"].items():
                    tags[key.strip()] = str(value)

            # Checking if this is a generation
            messages = inputs["messages"] if "messages" in inputs else []
            chain_type = self._get_chain_type(messages=messages)

            if chain_type == ChainType.TOOL:
                container, container_id = self._get_container(str(parent_run_id))
                if container is not None:
                    self.containers.set(
                        str(run_id),
                        f"{container}:{container_id}",
                        DEFAULT_TIMEOUT,
                    )

            if chain_type == ChainType.GENERATION:
                # Here we just create generation id and push it against the run_id
                # This we do to avoid any duplicate generations getting created
                # if users end up adding decorators or tracing the individual model calls

                # Checking if generation exists
                if self.generations.get(str(parent_run_id)) is not None:
                    # This means that the first message of the inputs is going to be AIMessage
                    # We will update this generation and end it
                    message = messages[-1]
                    generation_id = self.generations.get(str(parent_run_id))
                    if isinstance(message, AIMessage):
                        # Updating the generation
                        model = message.response_metadata.get("model", "unknown")
                        self.logger.generation_set_model(generation_id, model)
                        self.logger.generation_result(
                            generation_id=generation_id, result=message
                        )
                    return

                # This means that this is a new turn
                # We will check if the last message is AIMessage
                # If its AIMessage - then its most likely a tool call
                # Else we will pick the last HumanMessage and create a new generation
                if isinstance(messages[-1], AIMessage):
                    # This means its the next round
                    # Here checking if there is a tool_call
                    # If yes, updating chain
                    message = messages[-1]
                    tool_calls = message.tool_calls or []
                    if len(tool_calls) > 0:
                        container, container_id = self._get_container(
                            str(parent_run_id)
                        )
                        if container is not None:
                            self.containers.set(
                                str(run_id),
                                f"{container}:{container_id}",
                                DEFAULT_TIMEOUT,
                            )
                    # Rest of the flow will be handled in the tool_call_start and tool_call_end
                    return
                elif isinstance(messages[-1], ToolMessage):
                    # We attach the container again to send the events
                    container, container_id = self._get_container(str(parent_run_id))
                    if container is not None:
                        self.containers.set(
                            str(run_id),
                            f"{container}:{container_id}",
                            DEFAULT_TIMEOUT,
                        )
                else:
                    # We will only keep last message in messages
                    messages = [messages[-1]]

                span_id = self.chains.get(str(parent_run_id))
                # Creating a new generation
                generation_id = str(uuid4())
                # Creating the generation
                # This is a placeholder generation
                # Rest of the information will be added when chain ends
                parsed_messages = parse_langchain_messages([messages])
                self.logger.span_add_generation(
                    span_id,
                    GenerationConfig(
                        id=generation_id,
                        name=span_name,
                        # openai is the default provider we use for now
                        provider="unknown",
                        model="unknown",
                        messages=parsed_messages,
                        tags=tags,
                    ),
                )
                self.containers.set(str(run_id), f"span:{span_id}", DEFAULT_TIMEOUT)
                self.generations.set(str(run_id), generation_id, DEFAULT_TIMEOUT)
                return

            if chain_type == ChainType.CHAIN:
                # It's a span, so create one and move on
                if parent_run_id is None:
                    span_config = SpanConfig(id=span_id, name=span_name, tags=tags)
                    # this is the core id so add it to the span
                    self.chains.set(str(run_id), span_id, DEFAULT_TIMEOUT)
                    container, container_id = self._get_container_from_metadata(run_id)
                    event_id = str(uuid4())
                    if container == "span":
                        span = self.logger.span_add_sub_span(container_id, span_config)
                        self.logger.span_event(span.id, event_id, "start", tags)
                    elif container == "trace":
                        span = self.logger.trace_add_span(container_id, span_config)
                        self.logger.span_event(span.id, event_id, "start", tags)
                    else:
                        # Checking if I have current trace
                        if self.metadata is not None:
                            trace_id = self.metadata["trace_id"]
                            if trace_id is not None:
                                span = self.logger.trace_add_span(trace_id, span_config)
                                self.logger.span_event(span.id, event_id, "start", tags)
                        else:
                            # Creating trace locally
                            trace_name = None
                            trace = self.logger.trace(
                                TraceConfig(id=str(run_id), name=trace_name)
                            )
                            span = trace.span(span_config)
                            self.logger.span_event(span.id, event_id, "start", tags)
                else:
                    span_id = self.chains.get(str(parent_run_id))
                    new_span_id = str(uuid4())
                    span_config = SpanConfig(id=new_span_id, name=span_name, tags=tags)
                    self.chains.set(str(run_id), new_span_id, DEFAULT_TIMEOUT)
                    span = self.logger.span_add_sub_span(span_id, span_config)

                if span is not None:
                    # Getting existing metadata
                    new_meta = {"span_id": span.id}
                    run_meta = self.metadata_store.get(str(run_id))
                    if run_meta is not None:
                        new_meta.update(run_meta)
                    # Here we will set the container for the given run_id
                    self.containers.set(str(run_id), f"span:{span.id}", DEFAULT_TIMEOUT)
                    self.metadata_store.set(str(run_id), new_meta, DEFAULT_TIMEOUT)

                span_id = self.chains.get(str(parent_run_id))
                if span_id is None:
                    logger.error("[MaximSDK] No span id found for chain")
                    return
                tool_message: ToolMessage = next(
                    (
                        message
                        for message in reversed(messages)
                        if isinstance(message, ToolMessage)
                    ),
                    None,
                )
                if tool_message is None:
                    return
                # Find the AI message in messages
                # Which has the tool call
                args = ""
                for message in messages:
                    if isinstance(message, AIMessage):
                        tool_calls = message.tool_calls or []
                        for tool_call in tool_calls:
                            if tool_call["id"] == tool_message.tool_call_id:
                                args = tool_call["args"]
                                break
                tool_call_id = str(uuid4())
                tool_call = self.logger.span_tool_call(
                    span_id,
                    ToolCallConfig(
                        id=tool_call_id, name=tool_message.name, args=args, tags=tags
                    ),
                )
                print("tool start")
                return
        except Exception as e:
            logger.error(f"[MaximSDK] Failed to parse chain-start: {e}")

    def on_chain_end(self, outputs: Union[str, Dict[str, Any]], **kwargs: Any) -> Any:
        try:
            # We hide the hidden chains
            if "langsmith:hidden" in kwargs.get("tags", []):
                return
            run_id = kwargs.get("run_id", None)
            parent_run_id = kwargs.get("parent_run_id", None)
            tags = {
                tag.split(":")[0]: tag.split(":")[1]
                for tag in kwargs.get("tags", [])
                if ":" in tag
            }
            # Getting chain type
            chain_type = self._get_chain_type(
                messages=(
                    outputs["messages"] if isinstance(outputs, dict) else [outputs]
                ),
            )
            if chain_type == ChainType.CHAIN or chain_type == ChainType.TOOL:
                if isinstance(outputs, str):
                    # we fire this as event
                    if parent_run_id is not None:
                        container, container_id = self._get_container(parent_run_id)
                        event_id = str(uuid4())
                        if container == "span":
                            self.logger.span_event(
                                container_id, event_id, outputs, tags
                            )
                        elif container == "trace":
                            self.logger.trace_event(
                                container_id, event_id, outputs, tags
                            )
                if run_id is not None:
                    span_id = self.chains.get(str(run_id))
                    if span_id is not None:
                        self.logger.span_end(span_id)
                    self.chains.delete(str(run_id))
                    self.metadata_store.delete(str(run_id))

            if chain_type == ChainType.GENERATION:
                self.generations.delete(str(run_id))

            self.containers.delete(str(run_id))
        except Exception as e:
            logger.error(f"[MaximSDK] Failed to parse chain-end: {e}")

    def on_chain_error(
        self, error: Union[Exception, BaseException, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        print("chain-error", error, kwargs)

    def on_agent_action(
        self,
        action: AgentAction,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        print("on-agent-action", action, run_id, parent_run_id, kwargs)

    def on_agent_finish(
        self,
        finish: AgentFinish,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        print("on-agent-finish", finish, run_id, parent_run_id, kwargs)

    # Tool callbacks

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> Any:
        try:
            run_id = kwargs.get("run_id", None)
            parent_run_id = kwargs.get("parent_run_id", None)
            name = serialized.get("name", None)
            description = serialized.get("description", None)
            container, container_id = self._get_container(str(parent_run_id))
            if container is None:
                container, container_id = self._get_container_from_metadata(
                    parent_run_id
                )
            if container == "span":
                tool_call = self.logger.span_add_tool_call(
                    container_id,
                    ToolCallConfig(
                        id=str(uuid4()),
                        name=name,
                        description=description,
                        args=input_str,
                    ),
                )
                self.tool_calls.set(str(run_id), tool_call.id, DEFAULT_TIMEOUT)
            elif container == "trace":
                tool_call = self.logger.trace_add_tool_call(
                    container_id,
                    ToolCallConfig(
                        id=str(uuid4()),
                        name=name,
                        description=description,
                        args=input_str,
                    ),
                )
                self.tool_calls.set(str(run_id), tool_call.id, DEFAULT_TIMEOUT)
            else:
                logger.error(
                    "[MaximSDK] No container found for tool call. This is invalid state."
                )
        except Exception as e:
            logger.error(f"[MaximSDK] Failed to parse tool-start: {e}")

    def on_tool_end(self, output: Any, **kwargs: Any) -> Any:
        try:
            run_id = kwargs.get("run_id", None)
            tool_call_id = self.tool_calls.get(str(run_id))
            if tool_call_id is None:
                logger.error(
                    "[MaximSDK] No tool call id found for tool end. This is invalid state."
                )
                return
            if isinstance(output, ToolMessage):
                if output.status == "success":
                    self.logger.tool_call_result(tool_call_id, output.content)
                elif output.status == "error":
                    self.logger.tool_call_error(tool_call_id, output.content)
            else:
                logger.error(
                    f"[MaximSDK] Invalid output type {type(output)}. This is invalid state."
                )
        except Exception as e:
            logger.error(f"[MaximSDK] Failed to parse tool-end: {e}")

    def on_tool_error(
        self, error: Union[Exception, BaseException, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        try:
            run_id = kwargs.get("run_id", None)
            tool_call_id = self.tool_calls.get(str(run_id))
            if tool_call_id is None:
                logger.error(
                    "[MaximSDK] No tool call id found for tool end. This is invalid state."
                )
                return
            self.logger.tool_call_error(tool_call_id, str(error))
        except Exception as e:
            logger.error(f"[MaximSDK] Failed to parse tool-end: {e}")

    def on_text(self, text: str, **kwargs: Any) -> Any:
        print("on-text", text, kwargs)



import json
from dataclasses import asdict, dataclass
from typing import Dict, List, Optional, Union


@dataclass
class Folder():
    id: str
    name: str
    parentFolderId: str
    tags: Optional[Dict[str, Union[str, int, bool, None]]] = None


class FolderEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Folder):
            return asdict(obj)
        return super().default(obj)


@dataclass
class Error():
    message: str


@dataclass
class MaximFolderResponse():
    data: Folder
    error: Optional[Error] = None


@dataclass
class MaximFoldersResponse():
    data: List[Folder]
    error: Optional[Error] = None



import time
from collections import OrderedDict


class ExpiringKeyValueStore:
    def __init__(self):
        self.store = OrderedDict()
    
    def set(self, key, value, expiry_seconds):
        expiry_time = time.time() + expiry_seconds
        self.store[key] = (value, expiry_time)
        self._evict_expired()
    
    def get(self, key):
        if key in self.store:
            value, expiry_time = self.store[key]
            if time.time() < expiry_time:
                return value
            else:
                del self.store[key]
        return None
    
    def delete(self, key):
        if key in self.store:
            del self.store[key]
    
    def _evict_expired(self):
        current_time = time.time()
        expired_keys = [key for key, (_, expiry_time) in self.store.items() if current_time >= expiry_time]
        for key in expired_keys:
            del self.store[key]


from dataclasses import dataclass
from typing import Dict, Literal, Union


class QueryRuleType(str):
    DeploymentVar = "deploymentVar"
    Tag = "tag"


@dataclass
class QueryRule():
    query: str
    operator: Literal["AND", "OR"]
    exactMatch: bool
    scopes: Dict[str, str]


class QueryBuilder:
    def __init__(self):
        self.query: str = ""
        self.scopes: Dict[str, str] = {}
        self.operator: Literal["AND", "OR"] = "AND"
        self.isExactMatch: bool = False

    def and_(self) -> 'QueryBuilder':
        self.operator = "AND"
        return self

    def or_(self) -> 'QueryBuilder':
        self.operator = "OR"
        return self

    def folder(self, folderId: str) -> 'QueryBuilder':
        self.scopes["folder"] = folderId
        return self

    def exactMatch(self) -> 'QueryBuilder':
        self.isExactMatch = True
        return self

    def deploymentVar(self, key: str, value: Union[str, int, bool], enforce: bool = True) -> 'QueryBuilder':
        if len(self.query)>0:
            self.query += ","
        self.query += f"{'!!' if enforce else ''}{key}={value}"
        return self

    def tag(self, key: str, value: Union[str, int, bool], enforce: bool = False) -> 'QueryBuilder':
        if len(self.query)>0:
            self.query += ","
        self.query += f"{'!!' if enforce else ''}{key}={value}"
        return self

    def build(self) -> QueryRule:
        if len(self.query.strip())==0:
            raise ValueError("Cannot build an empty query. Please add at least one rule (deploymentVar or tag).")
        return QueryRule(
            query=self.query,
            operator=self.operator,
            exactMatch=self.isExactMatch,
            scopes=self.scopes
        )